/* csv2dat */
#define VERSION "4.11.2"

/* Uncomment the next line or use -DUNIX to compile on a unix system */
/* #define UNIX */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

void usage(void)
{
	fprintf(stderr,"Usage (v%s): csv2dat [-a] [-c|-d] [-q|-v] [-h|-?] [-r] <filename.CSV>\n",
		VERSION);
	fprintf(stderr,"\ta=ALL, c=CHECKED, d=VALIDATED, q=QUIET, v=VERBOSE, h|?=HELP\n");
}

/* Copyright (c) 2002-05 Rick Parsons.

	This software is published Open Source Freeware. You are free to use the
	programs either in full or part without charge. I would ask however that
	this notice is included with any extract or onward distribution.

	No charge may be made for the use of this code, however there is no
	objection to including modest extracts (with acknowledgement) in otherwise
	commercial software.

	No warranty is given or implied by the use of the software which you do at
	your own risk. No compensation can be considered regarding damage to data,
	computers or any thing else arising as a result of using these programs.
*/

#define TRUE 1
#define FALSE 0

int options(int, char **);
char * extract(char *, int *, int);
/* 4.5.0 */
void noquery(char *, int, char *);
/* 4.10.0 */
int cfind(char *, char *, int);
char * pad(char *, int);
int numeric(char *);
/* 4.9.1 */
int alphabetic(char *);
/* 4.11.0 */
int isvalid(unsigned char);
int isusual(char);
char part2(char *);
char * capitals(char *);
void inits(char *);
void splitpath(const char *, char *, char *, char *);
char * makepath(const char *, const char *, const char *);
char * fgetm(char *, int, FILE *);
int decode_census(char *);
/* 4.7.0 */
unsigned char tolower8(unsigned char);
unsigned char toupper8(unsigned char);

static FILE * err;
/* n = record number, year = census year, brief = mode,
	extract_err = error detected during extract of data from a field
	extract_content = if anything in a field
	rlen = record length (of output), verbose & quiet  = message verbosity
	nerr & nwarn = number of errors and warnings, sct = Scottish census
	ipart = non zero if partial piece
	rcheck = true if relation field is being checked against a list
	hold = a work area, date = todays date/time
*/
/* 4.10.0 add rcheck */
static int n, year, brief, extract_err, extract_content, rlen,
	verbose, quiet, nerr, nwarn, sct, ipart, rcheck;
/* Note, a lot of the character buffers are unsigned char but this is not
	declared unless necessary for the compiler */
static char hold[100];
static char date[40];

#define BUFSZ 1024

int _main(int argc, char *argv[])
{
	extern int n, brief, year, extract_err, extract_content, rlen,
		verbose, quiet, nerr, nwarn, sct, ipart;
	int i, nc, ppp, ppf, nage, maxhh, new_household, ret, empty, offset = 3,
		ccheck = FALSE, wls = FALSE;
	extern char hold[];
	extern char date[];
	char buf[BUFSZ];
	time_t tod;

	char del[2];
	/* 4.9.0 change regno fields to mean book number */
	char book[5], p_book[5], n_book[5];
	char household[7], p_household[7];
	char household_mem[5], p_household_mem[5];
	char parish[21], p_parish[21];
	char ed[5], p_ed[5], o_ed[5];
	char folio[6], p_folio[6], o_folio[6];
	char page[5], p_page[5];
	char sched[5], p_sched[5], o_sched[5];
	char house_num[6], p_house_num[6], o_house_num[6];
	char house_name[31], p_house_name[31];
	char buvn[2], p_buvn[2];
	char surname[25];
	char forenames[25];
	char name_err[2];
	char rel[7];
	char cond[2];
	char sex[2];
	char age[5];
	char detail_err[2];
	char occ[32]; /* one more than you would expect - see later */
	char emp[2];
	char occ_err[2];
	char chap[4];
	char place[21];
	char place_err[2];
	char dis[7];
	char welsh[2];
	char notes[45];
	char parishe[21], p_parishe[21];
	char altchap[4];
	char altplace[21];
	char ed_a, folio_a, sched_a, house_num_a, age_u, * cp;
	char addr_err[2];

	/* 4.7.3 extend all path and file names to 128 (XP likes long paths) */
	char input[128];
	char fname[128];
	char path[128];
	/* 4.7.4 ppath = program path - where this binary is located */
	char ppath[128];
	char temp[128];
	char ext[128];
	char output[128];
	char error[128];
	FILE * in;
	FILE * out;
	extern FILE * err;
	/* 4.6.0 pointer to county code table */
	char * ccode;
	/* 4.10.0 */
	char * rcode;

	/******************************/
	/* Decode the call parameters */
	/******************************/

	if (argc == 1)
	{
		/* If no arguments we call the usage routine and exit */
		usage();
		return 1;
	}

	/* handle the program options */
	i = options(argc,argv);

	/* If there is still no file name then what do we do? */
	if (i == 0)
		return 2;

	/* 4.7.4 get path of the binary for this program - the fname and ext are ignored */

	/* check size of path before using (csv2dat is only 7 characters, allow for 8) */
	if (strlen(argv[0]) > 126)
	{
		fprintf(stderr, "Fatal: Program path too long, move it closer to the root of the disk\n");
		return 12;
	}

	splitpath(argv[0], ppath, fname, ext);

	/* 4.7.3 check bounds of path before using */
	if (strlen(argv[i]) > 127)
	{
		fprintf(stderr, "Fatal: File path too long, move it closer to the root of the disk\n");
		return 12;
	}

	splitpath(argv[i], path, fname, ext);
	strcpy(input, argv[i]);

	/* Open the input and output files. Get the error file open asap. */

	strcpy(error, makepath(path, fname, "erc"));
	if ((err = fopen(error, "w")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Messages file (%s) failed to open\n", error);
		perror("Fatal");
		return 12;
	}

	if (*ext == (char)NULL)
	{
		fprintf(err, "Fatal: Full file name required including the three letter extension - %s\n", input);
		fprintf(stderr, "Fatal: Full file name required including the three letter extension\n");
		fclose(err);
		return 12;
	}

	/* decode the file extension to make sure it is the right type */
	if (strcmp(ext, "csv") != 0 && strcmp(ext, "CSV") != 0)
	{
		fprintf(err, "Fatal: File must be a .csv (comma separated values)\n");
		fprintf(stderr, "Fatal: File must be a .csv (comma separated values)\n");
		fclose(err);
		return 12;
	}

	/* decode the file name to work out what we are dealing with */

	ret = decode_census(fname);
	if (ret > 0)
	{
		fclose(err);
		return ret;
	}

	/* 4.6.0 read in the Chapman code vs. County name list (optional) */
	nwarn = 0;

	/* 4.9.0 moved here from later */
	nerr = 0;

	/* 4.6.4 only do it for years after 1841 */
	if (year > 1841)
	{
		/* 4.7.4 locate the aux file in the binary directory */
		/* 4.10.3 correct the case of the name */
		strcpy(temp, makepath(ppath, "CHAP1891", "TXT"));

	if ((in = fopen(temp, "r")) == (FILE *)NULL)
	{
		if (! quiet) /* 4.10.0 corrected */
			fprintf(err, "Warning: Aux file %s failed to open, county code checking disabled\n\n", temp), nwarn++;
	}
	else
	{
		/* 4.6.4 delay allocating memory till we know we will need it */
		ccode = (char *)malloc(sizeof(char) * 4 * 150);

		/* read lines of the form county[20]ccode[4], one per line */
		for (i = 0, cp = ccode; fgets(buf, 40, in) != (char *)NULL; i++, cp += 4)
		{
			/* ignore the county name for this program */
			strncpy(cp, buf + 20, 3);
			*(cp + 3) = (char)NULL;
		}
		fclose(in);

		if (i >= 150)
		{
			/* in practice, this is too late as we have already
			overrun the buffer */
			if (!quiet)
				fprintf(err, "Warning: Too many counties in aux file, county code checking disabled\n\n"), nwarn++;
		}
		else
		{
			*cp = (char)NULL;
			ccheck = TRUE;
		}
	}

		/* 4.10.0 locate a second aux file in the binary directory */
		strcpy(temp, makepath(ppath, "relcodes", "txt"));

	if (!rcheck || (in = fopen(temp, "r")) == (FILE *)NULL)
	{
		if (! quiet && rcheck)
			fprintf(err, "Warning: Aux file %s failed to open, relationship code checking disabled\n\n", temp), nwarn++;
		rcheck = FALSE;
	}
	else
	{
		rcode = (char *)malloc(sizeof(char) * 7 * 150);

		/* read lines of the form relcode[6], one per line */
		for (i = 0, cp = rcode; fgets(buf, 40, in) != (char *)NULL; i++, cp += 7)
		{
			*(strchr(buf, '\n')) = (char)NULL;
			strncpy(cp, buf, 6);
			*(cp + 6) = (char)NULL;
		}
		fclose(in);

		if (i >= 150)
		{
			/* in practice, this is too late as we have already
			overrun the buffer */
			if (!quiet)
				fprintf(err, "Warning: Too many relationships in aux file, relationship code checking disabled\n\n"), nwarn++;
			rcheck = FALSE;
		}
		else
			*cp = (char)NULL;
	}
	}

	/* Have to open the input in Binary to allow for MAC and Unix files */

	if ((in = fopen(input, "rb")) == (FILE *)NULL)
	{
		fprintf(err, "Fatal: Input file (%s) failed to open\n", input);
		fprintf(stderr, "Fatal: Input file (%s) failed to open\n", input);
		perror("Fatal");
		fclose(err);
		return 12;
	}

	/**********************************************/
	/* Verify what file type we have (BRIEF only) */
	/**********************************************/

	if (brief)
	{
		/* the first three rows are special
			piece, book, comment
			dummy (column names)
			dummy (column sizes)
		*/
		(void)fgetm(buf, BUFSZ, in);
		if (feof(in))
		{
			fprintf(err, "Fatal: row 1, No data in file, it appears to be empty\n");
			fprintf(stderr, "Fatal: No data in file, it appears to be empty\n");
			fclose(in);
			fclose(err);
			return 12;
		}


		if ((cp = strchr(buf, '\n')) != (char *)NULL)
			*cp = (char)NULL;

		i = 0; n = 1; /* for error messages if necessary */

		/* 4.6.3 increase size from 12 to 30 to allow for 1841 parish names etc. */
		cp = extract(buf, &i, 30);
		(void)capitals(cp);
		if (! quiet && strncmp(cp, "RG", 2) != 0 && strncmp(cp, "HO", 2) != 0 &&
			strncmp(cp, "RS", 2) != 0 && strncmp(cp, "HS", 2) != 0 &&
			strncmp(cp, "SCT", 3) != 0)
			/* extra new line because this warning comes out before the titles */
			fprintf(err, "Warning: Cell A1 doesn't contain a valid piece number\n\n"),
				++nwarn;

		/* 4.9.0 change regno fields to mean book number */
		strcpy(book, extract(buf, &i, 4));

		/* 4.9.0 if book number blank, set to zero */
		/* 4.11.1 report inconsistencies */
		if (year == 1841)
		{
			if (!sct)
			{
			/* Test for a non numeric book number field */
			/* 4.11.2 Non-numeric is now only a warning */
			if (!quiet && !numeric(book))
				/* extra new line because this message comes out before the titles */
				fprintf(err, "Warning: cell B1 contains a non-numeric book number %s\n\n", book), ++nwarn;

			if (*book == (char)NULL)
				strcpy(book, "0");
			}
			else /* wipe it out */
			{
				if (*book != (char)NULL && !quiet)
					/* extra new line because this warning comes out before the titles */
					fprintf(err, "Warning: ENG 1841 book number %s in cell B1 ignored for SCT piece\n\n", book),
						++nwarn;
				*book = (char)NULL;
			}
		}
		else /* wipe it out */
		{
			if (*book != (char)NULL && !quiet)
				/* extra new line because this warning comes out before the titles */
				fprintf(err, "Warning: 1841 book number %s in cell B1 ignored for %d piece\n\n", book, year),
					++nwarn;
			*book = (char)NULL;
		}

		(void)fgetm(buf, BUFSZ, in);
		if (feof(in))
		{
			fprintf(err, "Fatal: row 2, No data in file, only one header row was found\n");
			fprintf(stderr, "Fatal: No data in file, only one header row was found\n");
			fclose(in);
			fclose(err);
			return 12;
		}

		i = 0; n = 2; /* for error messages if necessary */

		/* count the number of items on the line */
		for (nc = 0; nc < 30 ; nc++)
		{
			/* most errors are ignored */
			(void)extract(buf, &i, 50);
			if (*hold == (char)NULL)
				break;
		}

		/* 4.9.0 Can be one extra column for 1841 now */
		/* DAT = 25, CHE = 26, VLD = 28 or 29 */
		if (! ((nc == 25 && rlen == 256) || (nc == 26 && rlen == 276) ||
			((nc == 28 || nc == 29) && rlen == 299)))
		{
			fprintf(err,
				"Fatal: row 2, File type switch does not match number of columns in data (%d)\n\tall the titles in row 2 must be non-blank\n", nc);
			fprintf(stderr,
				"Fatal: File type switch does not match number of columns in data\n\tall the headers in row 2 must be non-blank\n");
			fclose(in);
			fclose(err);
			return 12;
		}
		/* alternatively at this point we could abandon the switches and
			work out the file type from this value, however there is no
			obvious way to do it for ALL format */

		/* lose the next line */
		(void)fgetm(buf, BUFSZ, in);
		if (feof(in))
		{
			fprintf(err, "Fatal: row 3, No data in file, only two header rows were found\n");
			fprintf(stderr, "Fatal: No data in file, only two header rows were found\n");
			fclose(in);
			fclose(err);
			return 12;
		}
	}

	/* open the output file now that there is a chance we may put something in it */
	strcpy(output, makepath(path, fname,
		rlen == 299 ? "vld" : (rlen == 256 ? "dat" : "val")));

	if ((out = fopen(output, "wb")) == (FILE *)NULL)
	{
		fprintf(err, "Fatal: Output file (%s) failed to open\n", output);
		fprintf(stderr, "Fatal: Output file (%s) failed to open\n", output);
		perror("Fatal");
		fclose(in);
		fclose(err);
		return 12;
	}

	/****************************************/
	/* preset some values for counters etc. */
	/****************************************/

	if ((tod = time((time_t *)NULL)) != -1)
	{
		strcpy(date, asctime(gmtime(&tod)));
		/* get rid of the newline */
		i = strlen(date);
		date[i - 1] = (char)NULL;
		strcat(date, " UTC");
	}
	else
		*date = (char)NULL;

	fprintf(err, "Error listing for csv2dat (Rick Parsons) v%s %s converting %s\n"
"[When reporting program errors, be sure to send this file and the original data file]\n\n",
		VERSION, date, input);
	fprintf(err,
		"Census year = %d, input format = %s, output format = %s%s\n\n",
		year, brief ? "SS-CENS" : "ALL",
		(rlen == 299) ? "VALD-REV" : ((rlen == 256) ? "IN-CENS" : "CHECK-CENS"),
		(quiet) ? ", QUIET" :((verbose) ? ", VERBOSE" : ""));

	n = 0; nc = 1;
	/* 4.9.0 moved earlier
	nerr = 0;
*/

	if (brief)
	{
		fprintf(err,
"Subtract %d from row numbers in messages to get the corresponding record in the output file.\n\n", offset);

		strcpy(household, "0");
		strcpy(household_mem, "0");

		n = 3; nc = 4; /* nc is the number of the first data row */
		ppp = 0; ppf = 0; /* people-per-page and page-per-folio counters */

		/* these are meant to break it if they are not over-ridden */
		strcpy(p_parish, "XX");
		strcpy(p_folio, "9999");
		strcpy(p_page, "999");
		strcpy(p_ed, "99");
		/* schedule numbers are not available for 1841 but we are using their
			presence to detect the start of a household. They should always be 0 */
		strcpy(p_sched, "999");

		/* Scotland uses the folios field for their book numbers. Set the default to zero.
		Yes, I know the terminology is confusing but these are not the same as the book numbers
		used on ENG/WLS 1841 pieces */
		if (sct)
			strcpy(p_folio, "0");

		/* these are valid default values */
		strcpy(p_parishe, "-");
		strcpy(p_house_num, "");
		strcpy(p_house_name, "-");
		strcpy(p_buvn, "-");

		/* 4.9.0 Preset the previous book number to the one read from cell B1 */
		strcpy(p_book, book);
	}
	else
	{
		/* give ALL mode a starting point for the sequence checks */
		strcpy(p_household, "0");
		strcpy(p_household_mem, "0");
		maxhh = 0;
	}

	/* 4.9.0 see if book numbers are in use */
	if (!quiet && year == 1841 && !sct)
	{
		if (atoi(book) > 100)
			fprintf(err, "Warning: Old format file, Book numbers not in use\n"),
				nwarn++;
		else if (atoi(book) == 0)
			fprintf(err, "Warning: Book numbers not set\n"), nwarn++;
	}

	/**************************************/
	/* big loop reading one row at a time */
	/**************************************/

	while (TRUE)
	{	/* the test for end of file moved here to let records through
			which have no newline on the end. */
		if (feof(in))
			break;

		(void)fgetm(buf, BUFSZ, in);

		if ((cp = strchr(buf, '\n')) != (char *)NULL)
			*cp = (char)NULL;

		/* now that the eof check is before the fget we get a blank line
			on the end of most files, so do it again with some other checks */
		if (feof(in) && *buf == (char)NULL)
			break;

		/* running total so the punter can see what is happening */
		n++;
		fprintf(stderr, "%d\r", n);

		i = 0;
		strcpy(del, " ");

		/**************************************/
		/* extract the data from the CSV line */
		/**************************************/

		extract_content = FALSE; /* reset so we can see if there is anything there */

		if (! brief)
		{
			strcpy(del, extract(buf, &i, 1));
			/* 4.9.0 change regno fields to mean book number */
			strcpy(book, extract(buf, &i, 4));
			strcpy(household, extract(buf, &i, 6));
			strcpy(household_mem, extract(buf, &i, 4));
		}

		strcpy(parish, extract(buf, &i, 20));
		/* 4.5.0 */
		noquery(parish, FALSE, (char *) NULL);

		if (brief && rlen > 256)
		{
			strcpy(parishe, extract(buf, &i, 20));
			noquery(parishe, FALSE, (char *) NULL);
		}

		strcpy(ed, extract(buf, &i, 4));
		strcpy(folio, extract(buf, &i, 5));
		strcpy(page, extract(buf, &i, 4));
		strcpy(sched, extract(buf, &i, 4));

		extract_err = FALSE; /* it is ignored up to this point */

		strcpy(house_num, extract(buf, &i, 5));
		strcpy(house_name, extract(buf, &i, 30));
		strcpy(buvn, extract(buf, &i, 1));

		addr_err[0] = (char)NULL;
		/* 4.5.0 */
		noquery(house_name, TRUE, addr_err);
		if (extract_err)
		{
			/* over-ride what was there!? - bad database design,
			This would foul up the remainder of the record if it previoulsly
			had an uninhabited type value so we save it just in case */
			strcpy(addr_err, "x");
			extract_err = FALSE;
		}

		strcpy(surname, extract(buf, &i, 24));
		strcpy(forenames, extract(buf, &i, 24));
		strcpy(name_err, extract(buf, &i, 1));
		/* 4.5.0 */
		noquery(surname, TRUE, name_err);
		noquery(forenames, TRUE, name_err);
		if (extract_err)
		{
			strcpy(name_err, "x");
			extract_err = FALSE;
		}

		strcpy(rel, extract(buf, &i, 6));
		strcpy(cond, extract(buf, &i, 1));
		strcpy(sex, extract(buf, &i, 1));
		strcpy(age, extract(buf, &i, 4));
		strcpy(detail_err, extract(buf, &i, 1));
		/* 4.5.0 */
		noquery(rel, TRUE, detail_err);
		if (extract_err)
		{
			strcpy(detail_err, "x");
			extract_err = FALSE;
		}

		strcpy(occ, extract(buf, &i, 30));

		if (brief)
			strcpy(emp, extract(buf, &i, 1));

		strcpy(occ_err, extract(buf, &i, 1));
		/* 4.5.0 */
		noquery(occ, TRUE, occ_err);
		if (extract_err)
		{
			strcpy(occ_err, "x");
			extract_err = FALSE;
		}

		strcpy(chap, extract(buf, &i, 3));
		strcpy(place, extract(buf, &i, 20));
		strcpy(place_err, extract(buf, &i, 1));

		noquery(place, TRUE, place_err);
		if (extract_err)
		{
			strcpy(place_err, "x");
			extract_err = FALSE;
		}

		if (brief && rlen > 276)
		{
			strcpy(altchap, extract(buf, &i, 3));
			strcpy(altplace, extract(buf, &i, 20));
			/* 4.5.2  also test extract_err */
			noquery(altplace, TRUE, place_err);
			if (extract_err)
			{
				strcpy(place_err, "x");
				extract_err = FALSE;
			}
		}

		strcpy(dis, extract(buf, &i, 6));
		/* 4.5.0 */
		noquery(place, FALSE, (char *) NULL);

		strcpy(welsh, extract(buf, &i, 1));
		strcpy(notes, extract(buf, &i, 44));
		/* don't strip things off the notes */

		if (! brief && rlen > 256)
		{
			strcpy(parishe, extract(buf, &i, 20));
			/* 4.5.0 */
			noquery(parishe, FALSE, (char *) NULL);
		}

		if (! brief && rlen > 276)
		{
			strcpy(altchap, extract(buf, &i, 3));
			strcpy(altplace, extract(buf, &i, 20));
			/* 4.5.2 also test extract_err */
			noquery(altplace, TRUE, place_err);
			if (extract_err)
			{
				strcpy(place_err, "x");
				extract_err = FALSE;
			}
		}

		/* 4.9.0 Extract the possible book number field */
		if (brief && rlen > 276 && year == 1841 && !sct)
			strcpy(book, extract(buf, &i, 4));

		if (! extract_content)
		{
			if (! quiet)
				fprintf(err, "Warning: row %d blank, ignored\n", n), nwarn++;
			offset++;
			continue;
		}

		/*************************************************************/
		/* Sort out most of the differences between BRIEF & ALL mode */
		/*************************************************************/

		/* assume we are continuing the previous household for the moment */
		new_household = FALSE;

		if (brief)
		{
			sprintf(household_mem, "%d", atoi(household_mem) + 1);

			/* first 4 (5) columns always fill-down if required */
			if (*parish == (char)NULL)
				strcpy(parish, p_parish);
			if (rlen > 256 && *parishe == (char)NULL)
				strcpy(parishe, p_parishe);
			if (*ed == (char)NULL)
				strcpy(ed, p_ed);
			if (*folio == (char)NULL)
				strcpy(folio, p_folio);
			if (*page == (char)NULL)
				strcpy(page, p_page);

			/* next 4 columns only fill down if sched hasn't been set */
			if (*sched == (char)NULL)
			{
				strcpy(sched, p_sched);
				/* defer the house number/name until later */
				if (*buvn == (char)NULL)
					strcpy(buvn, p_buvn);
			}
			else
			{	/* increment the household number */
				sprintf(household, "%d", atoi(household) + 1);
				/* reset the household member number */
				strcpy(household_mem, "1");
				new_household = TRUE;
			}
			/* the rest of the record is always in full (no fill down) */

			if (strcmp(notes, "DELETED") == 0)
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, is marked DELETED\n", n), nwarn++;
				strcpy(del, "D");
			}
		}
		else
		{
			/* check validity of the first values */
			if (household[0] == (char)NULL || ! numeric(household))
				fprintf(err,
					"Error: row %d, household number = %s not numeric\n",
					n, household), nerr++;
			else if (atoi(household) == 0)
				fprintf(err,
					"Error: row %d, household number = 0 not valid\n",
					n), nerr++;

			if (! quiet && household_mem[0] == (char)NULL)
				fprintf(err,
					"Warning: row %d, household member number blank\n",
					n), nwarn++;
			else if (! numeric(household_mem))
				fprintf(err,
					"Error: row %d, household member number = %s not numeric\n",
					n, household_mem), nerr++;
			else if (! quiet && atoi(household_mem) == 0)
				fprintf(err,
					"Warning: row %d, household member number = 0 not valid\n",
					n), nwarn++;

			/* check that the household numbering makes sense */
			if (strcmp(household, p_household) != 0)
			{	/* new household */
				strcpy(p_household_mem, "0"); /* start numbering members again */
				if (verbose && atoi(household) != atoi(p_household) + 1)
					fprintf(err,
						"Info: row %d, household number = %s not sequential\n",
						n, household);
				strcpy(p_household, household);

				new_household = TRUE;
			}
			else if (! quiet && atoi(household_mem) != atoi(p_household_mem) + 1)
				fprintf(err,
					"Warning: row %d, household member number = %s not sequential\n",
					n, household_mem), nwarn++;
			strcpy(p_household_mem, household_mem);
		}

		/* we should now have a full record in both modes except for the
			address fields. Some errors are only relevant to one
			mode, but many, though superfluous, are checked anyway */

		/* extract the optional alphabetic parts, but save the originals first */
		strcpy(o_ed, ed); ed_a = part2(ed);
		strcpy(o_folio, folio); folio_a = part2(folio);

		/* 4.4.10 Check the length of the remaining field */
		if (strlen(ed) > 3)
		{
			fprintf(err, "Error: row %d, ED > 3 digits\n", n), nerr++;
			strcpy(ed, "999");
		}
		if (strlen(folio) > 4)
		{
			fprintf(err, "Error: row %d, folio/SCT book > 4 digits\n", n), nerr++;
			strcpy(folio, "9999");
		}

		/* it is possible that this will leave us with nothing (because of # codes) */
		if (*ed == (char)NULL)
			strcpy(ed, "0");

		/**********************************************/
		/* checks applicable to the first record only */
		/**********************************************/

		/* Some of these could be supressed for partial pieces */

		if (n == nc)
		{
			/* 4.9.0 no longer relevant
			if (! brief && verbose && book[0] == (char)NULL)
				fprintf(err, "Info: row %d, %s missing\n", n,
					(year == 1841) ? "parish number": "registration district");
			*/
			if (! quiet && atoi(household) != 1)
				fprintf(err, "Warning: row %d, initial household number = %s\n",
					n, household), nwarn++;

			if (! quiet && atoi(household_mem) != 1)
				fprintf(err, "Warning: row %d, initial household member number = %s\n",
					n, household_mem), nwarn++;

			/* also make sure that for BRIEF format, that all essential data was
				included on the first line */
			if (strcmp(parish, "XX") == 0)
				fprintf(err, "Error: row %d, initial civil parish name not set\n",
					n), nerr++;

			if (strcmp(ed, "99") == 0)
				fprintf(err, "Error: row %d, initial ED not set\n", n), nerr++;
			else
				strcpy(p_ed, "0");	/* so that sequence check works */

			/* This will never happen with SCT because the default is zero */
			if (strcmp(folio, "9999") == 0)
				fprintf(err, "Error: row %d, initial folio number not set\n", n),
					nerr++;
			else
				strcpy(p_folio, "0");	/* so that sequence check works */

			if (strcmp(page, "999") == 0)
				fprintf(err, "Error: row %d, initial page number not set\n", n),
					nerr++;

			/* this is independant of other schedule tests, which come later */
			if (strcmp(sched, "999") == 0)
				fprintf(err, "Error: row %d, initial schedule number not set\n",
					n), nerr++;

			/* 4.8.6 For the first record ony, set the Welsh flag iff the language
			field is non blank 4.8.8 but not scottish pieces! */
			if (! sct && year >= 1881 && welsh[0] != (char)NULL)
				wls = TRUE;

		}
		/* 4.9.0 no longer relevant
		else if (strcmp(p_book, book) != 0)
				fprintf(err, "Error: row %d, registration district changed\n", n),
					nerr++;
		*/

		/********************/
		/* main data checks */
		/********************/


		if (ed[0] == (char)NULL || ! numeric(ed))
		{
			fprintf(err, "Error: row %d, ED = %s not numeric\n", n, ed), nerr++;
			/* force it to 99 so we don't get repeated errors in BRIEF mode */
			strcpy(ed, "99");
		}

		if (ed_a != ' ' && ! isalnum(ed_a))
		{
			fprintf(err, "Error: row %d, ED suffix = %c not alphanumeric\n",
				n, ed_a), nerr++;
			ed_a = ' ';
		}

		if (folio[0] == (char)NULL || ! numeric(folio))
		{
			fprintf(err, "Error: row %d, folio/SCT book = %s not numeric\n", n, folio),
				nerr++;
			/* force it to 9999 so we don't get repeat errors in BRIEF mode */
			strcpy(folio, "9999");
		}

		if (folio_a != ' ' && ! isalpha(folio_a))
		{
			fprintf(err, "Error: row %d, folio/SCT book suffix = %c not alphabetic\n",
				n, folio_a), nerr++;
			folio_a = ' ';
		}

		if (page[0] == (char)NULL || ! numeric(page))
		{
			fprintf(err, "Error: row %d, page number = %s not numeric\n", n, page),
				nerr++;
			/* force it to 999 so we don't get repeated errors in BRIEF mode */
			strcpy(page, "999");
		}

		/* 4.8.9 Convert dummy 'Z' del markers to blank. These were put in by
			dat2csv to stop columns vanishing */
		if (del[0] == 'Z')
			del[0] = ' ';

		/* convert any empty fields to the required default and set the case right
			4.6.2 move earlier */
		if (del[0] == (char)NULL)
			del[0] = ' ';

		/* 4.9.1 Test for unusual characters */
		if (!quiet && ! alphabetic(parish))
			fprintf(err, "Warning: row %d, parish = %s contains unusual characters\n",
				n, parish), nwarn++;

		/* This can only happen in ALL mode */
		if (parish[0] == (char)NULL)
			fprintf(err, "Error: row %d, civil parish name missing\n", n), nerr++;
		else
			inits(parish);

		/* 4.8.11 Reinstate Eccl. Districts in 1851 */
		if (year > 1841 && rlen > 256)
		{
			/* 4.9.1 Test for unusual characters */
			if (!quiet && ! alphabetic(parishe))
				fprintf(err, "Warning: row %d, Eccl. District = %s contains unusual characters\n",
					n, parishe), nwarn++;

			if (parishe[0] == (char)NULL)
			{
				/* This can only happen in ALL mode */
				if (! quiet)
					fprintf(err, "Warning: row %d, Eccl. District name missing\n",
						n), nwarn++;
				strcpy(parishe, "-");
			}
			else
				inits(parishe);
		}

		/* 4.9.0 Handle the 1841 ENG/WLS book numbers */
		if (year == 1841 && !sct)
		{
			if (brief)
			{
				/* 4.11.2 Non-numeric is now only a warning */
				if (!quiet && !numeric(book))
					fprintf(err, "Warning: row %d, book number = %s not numeric\n", n, book),
						nwarn++;

				/* if the folio numbers go backwards,
					this is a trigger to increment the book number.
					It is the previous one we increment, not the one just read */
				strcpy(n_book, p_book);
				if (atoi(folio) < atoi(p_folio) && atoi(n_book) != 0)
					sprintf(n_book, "%d", atoi(n_book) + 1);

				/* The book number field can now be set if required */
				if (*book == (char)NULL || strcmp(book, "0") == 0)
					strcpy(book, n_book);
			}

			if (! quiet && atoi(book) != 0 &&
				atoi(book) != atoi(p_book) && atoi(book) != atoi(p_book) + 1)
				fprintf(err, "Warning: row %d, Book number = %s not sequential\n",
					n, book), nwarn++;
		}

		/* check (but not enforce) that ED numbers are in sequence */
		if (strcmp(o_ed, p_ed) != 0) /* ED has changed */
		{
/* SANITY ALERT - atoi doesn't care about non numeric stuff after the number */
/* 4.8.3 ENG/WLS 1841 whole pieces can have non sequential ED numbers */
			if (! quiet && (sct || year != 1841 || ipart != 0) &&
				atoi(ed) != 0 && atoi(p_ed) != 0 &&
				atoi(ed) != atoi(p_ed) && atoi(ed) != atoi(p_ed) + 1)
				fprintf(err, "Warning: row %d, ED = %s not sequential\n",
					n, o_ed), nwarn++;

			if (! quiet && ed_a == ' ' && atoi(page) > 1)
				fprintf(err, "Warning: row %d, first page of ED = %s\n",
					n, page), nwarn++;

			if (! quiet && atoi(folio) == 0)
				fprintf(err, "Warning: row %d, folio/SCT book = 0\n", n), nwarn++;

			/* SCT does not have folios, their book numbers don't require this check
			4.8.4 nor do part pieces */
			if ( ! sct && ipart == 0)
			{
				if (! quiet && atoi(folio) > atoi(p_folio) + 5)
					fprintf(err,
						"Warning: row %d, excessive gap before folio %s\n",
						n, o_folio), nwarn++;
			}

			/* 4.6.2 change of ED, reset line and page counts */
			ppp = 1; ppf = 1;
		}
		else /* we (think we) are in the same ED */
		{	/* check ascending page numbers.
				There may be some spurious messages in 1841 */
			if (! quiet && atoi(page) != atoi(p_page) && atoi(page) != atoi(p_page) + 1)
				fprintf(err,
					"Warning: row %d, page number = %s not sequential\n",
					n, page), nwarn++;

			/* SCT does not have folios, their book numbers don't require this check
				There may be some spurious messages in 1841 */
			if (! quiet && ! sct && atoi(folio) != atoi(p_folio)
				&& atoi(folio) != atoi(p_folio) + 1)
				fprintf(err, "Warning: row %d, folio number = %s not sequential\n",
					n, o_folio), nwarn++;

		/* 4.6.2 moved inside of new ED check (sorry about indent) */
		if (strcmp(page, p_page) != 0)
		{
			ppp = 0;
			/* SCT does not have folios, their book numbers don't require this check */
			if ( ! sct)
			{
				if (strcmp(o_folio, p_folio) != 0)
					ppf = 1;
				else if (++ppf > 2 && atoi(folio) != 0)
				{
					if (! quiet)
						fprintf(err,
							"Warning: row %d, too many pages on folio %s\n",
							n, o_folio), nwarn++;
					ppf = 0;
				}
			}
		}
		else if (ed_a != '8') /* not appropriate for shipping pages */
		{

			if (! quiet && strcmp(parish, p_parish) != 0)
				fprintf(err,
					"Warning: row %d, civil parish has changed in the middle of a page\n",
					n), nwarn++;

			/* 4.8.11 re-instate Eccl. Districts in 1851 */
			if (! quiet && year > 1841 && rlen > 256 && strcmp(parishe, p_parishe) != 0)
				fprintf(err,
					"Warning: row %d, Eccl. District has changed in the middle of a page\n",
					n), nwarn++;

 			if (! quiet && strcmp(o_folio, p_folio) != 0)
				fprintf(err,
					"Warning: row %d, folio/SCT book has changed in the middle of a page\n",
					n), nwarn++;

			/* 4.9.0 Check for book numbers changing in odd places */
			if (! quiet && year == 1841 && !sct && strcmp(book, p_book) != 0)
				fprintf(err,
					"Warning: row %d, Book number has changed in the middle of a page\n",
					n), nwarn++;

			if (++ppp > 31 && atoi(page) != 0)
			{
				/* this counts all rows but in fact multiple uninhabiteds are
					expanded by us so this can give spurious warnings. We
					ought to supress them. */
				if (! quiet)
					fprintf(err, "Warning: row %d, possibly too many lines on page %s\n",
						n, page), nwarn++;
				ppp = 0;
			}
		}
		}

		if (buvn[0] == (char)NULL)
			strcpy(buvn, "-");
		else
			buvn[0] = tolower(buvn[0]);

		/********************************/
		/* Uninhabited building records */
		/********************************/

		if (buvn[0] == 'b' || buvn[0] == 'u' || buvn[0] == 'v' || buvn[0] == 'n')
		{
			empty = TRUE;
			/* over-ride what was there so we get consistent values */
			if (buvn[0] == 'u')
				strcpy(surname, "Uninhabited");
			else if (buvn[0] == 'b')
				strcpy(surname, "Building in Progress");
			else if (buvn[0] == 'n')
				strcpy(surname, "Schedule not used");
			else
				strcpy(surname, "Visiting");

			/* blank out the rest of the record */
			forenames[0] = (char)NULL;
			strcpy(name_err, "-");
			rel[0] = (char)NULL;
			strcpy(cond, " ");
			strcpy(sex, " ");
			age[0] = (char)NULL;
			age_u = ' ';
			strcpy(detail_err, "-");
			occ[0] = (char)NULL;
			strcpy(occ_err, "-");
			chap[0] = (char)NULL;
			place[0] = (char)NULL;
			strcpy(place_err, "-");
			dis[0] = (char)NULL;
			strcpy(welsh, " ");
			/* 4.4.8 allow notes in these records notes[0] = (char)NULL; */

			/* uninhabiteds should always have a schedule number */
			if (! new_household)
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, uninhabited flag = %s but not new household\n",
						n, buvn), nwarn++;
				new_household = TRUE;
				/* we need to correct other things if this has happened */
				/* set the scheule - very probably should be 0 */
				strcpy(sched, "0");
				/* increment the household number */
				if (brief)
					sprintf(household, "%d", atoi(household) + 1);
				else
					/* keep it well clear of existing numbers */
					sprintf(household, "%d", atoi(household) + 10000);
				/* reset the household member number */
				strcpy(household_mem, "1");
			}
		}
		else
		{
			empty = FALSE;
			if (buvn[0] != 'x' && buvn[0] != '-')
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, uninhabited flag = %s invalid value\n",
						n, buvn), nwarn++;
				strcpy(buvn, "x");
			}

			/* this is an error we may have picked up earlier. It is now
				safe to copy it in. */
			if (addr_err[0] == 'x')
				strcpy(buvn, addr_err);

			if ((p_buvn[0] == 'b' || p_buvn[0] == 'u' || p_buvn[0] == 'v' ||
				p_buvn[0] == 'n') && ! new_household)
			{
				fprintf(err,
					"Error: row %d, the first record after an uninhabited building must be a new household\n",
					n), nerr++;
				new_household = TRUE;
				/* we need to correct other things if this has happened */

				/* we have no idea what the schedule number should be -
					This is why it is an Error not a Warning. Bob
					suggested set it to zero anyway and also set the
					error flag which I hadn't thought of. */
				strcpy(sched, "0");
				strcpy(buvn, "x");

				/* increment the household number */
				if (brief)
					sprintf(household, "%d", atoi(household) + 1);
				else
					/* keep it well clear of existing numbers */
					sprintf(household, "%d", atoi(household) + 10000);
				/* reset the household member number */
				strcpy(household_mem, "1");
			}
		}

		if (atoi(household) > maxhh)
			maxhh = atoi(household);

		/* this fill down was defered from earlier in case the new
			household recognition was wrong */
		if (brief && ! new_household)
		{
			if (*house_num == (char)NULL)
				strcpy(house_num, p_house_num);
			if (*house_name == (char)NULL)
				strcpy(house_name, p_house_name);
		}

		/* extract the remaining optional alpahbetic parts */
		strcpy(o_sched, sched); sched_a = part2(sched);
		strcpy(o_house_num, house_num); house_num_a = part2(house_num);

		/* 4.4.11 convert the ! if it is present so the correct number is used */
		/* 4.9.1 There is a bug here ... if sched is not numeric ... */
		if (sched_a == '!')
		{
			sprintf(sched, "%d", atoi(sched) + 1000);
			sched_a = ' ';
		}

		/* 4.4.10 Check the length of the remaining field
			Sched can be left until later */
		if (strlen(house_num) > 4)
		{
			fprintf(err, "Error: row %d, house number > 4 digits\n", n), nerr++;
			strcpy(house_num, "9999");
		}

		/**************************/
		/* check schedule numbers */
		/**************************/

		if (! numeric(sched))
		{
			fprintf(err,
				"Error: row %d, schedule number = %s not numeric\n", n, sched),
				nerr++;
			/* force it to 999 so we don't get repeated errors in BRIEF mode */
			strcpy(sched, "999");
		}

		if (sched_a != ' ' && ! isalpha(sched_a))
		{
			fprintf(err,
				"Error: row %d, schedule suffix = %c not alphabetic\n",
				n, sched_a), nerr++;
			sched_a = ' ';
		}

		if (strcmp(o_ed, p_ed) != 0) /* 4.4.11 (was o_sched) */
		{
			if (! quiet && ed_a == ' ' && year > 1841 && atoi(sched) > 1)
				fprintf(err, "Warning: row %d, first schedule of ED = %s\n",
					n, sched), nwarn++;
		}
		else
		{
			/* ALL mode only - this is unlikely to occur except by manual editing */
			if (! new_household && strcmp(o_sched, p_sched) != 0)
				fprintf(err,
					"Error: row %d, schedule number changed in middle of household\n",
					n), nerr++;
			/* note that the contrary - consecutive households with the same (!0) schedule
				number should only occur if they are uninhabited. */
			if (! quiet && ! empty && new_household && atoi(sched) != 0 &&
				strcmp(o_sched, p_sched) == 0)
				fprintf(err,
					"Warning: row %d, consecutive schedule numbers (%s) are the same\n",
					n, sched), nwarn++;

			/* 4.4.11 Not sure if this is correct for p_sched > 999? */
			if (! quiet && ed_a != '8' && atoi(o_sched) != atoi(p_sched) && atoi(sched) != 0 &&
				atoi(p_sched) != 0 && atoi(o_sched) != atoi(p_sched) + 1)
				fprintf(err,
					"Warning: row %d, schedule number = %s not sequential\n",
					n, sched), nwarn++;
				/* supresssed for merchant shipping which are often all over the place */
		}

		/* for 1841 files, we don't need schedules anymore so we can erase it now .
			Note that o_sched is preserved to keep sequence checks happy. */
		if (year == 1841)
			sched[0] = (char)NULL;

		/************************/
		/* check address fields */
		/************************/

		if (! numeric(house_num))
			fprintf(err, "Error: row %d, house number = %s not numeric\n",
				n, house_num), nerr++;

		if (house_name[0] == (char)NULL)
		{
			if (verbose && buvn[0] != 'x')
				fprintf(err, "Info: row %d, address missing\n", n);
			strcpy(house_name, "-");
			/* strcpy(buvn, "x"); /* This creates havoc with other tests later */
		}
		else
			inits(house_name);

		if (! new_household)
		{
			if (! quiet && strcmp(o_house_num, p_house_num) != 0)
				fprintf(err,
					"Warning: row %d, house number has changed in the middle of a household\n",
					n), nwarn++;
			if (strcmp(house_name, p_house_name) != 0)
			{	/* allow for imaginary changes where it is only a dash */
				if (strcmp(house_name, "-") == 0)
					strcpy(house_name, p_house_name);
				else if (! quiet)
					fprintf(err,
						"Warning: row %d, address has changed in the middle of a household\n",
						n), nwarn++;
			}
		}

		/***************************/
		/* personal detail records */
		/***************************/

		/* if (buvn[0] == '-' || buvn[0] == 'x') */
		if (! empty)
		{
			/**************/
			/* name group */
			/**************/

			if (name_err[0] == (char)NULL)
				strcpy(name_err, "-");
			else
				name_err[0] = tolower(name_err[0]);

			/* 4.9.1 Test for unusual characters */
			if (!quiet && ! alphabetic(surname))
				fprintf(err, "Warning: row %d, surname = %s contains unusual characters\n",
					n, surname), nwarn++;

			if (surname[0] == (char)NULL)
			{
				if (name_err[0] != 'x')
				{
					if (! quiet)
						fprintf(err, "Warning: row %d, surname missing\n",
							n), nwarn++;
					strcpy(name_err, "x");
				}
				strcpy(surname, "-");
			}
			else
				(void)capitals(surname);

			/* 4.9.1 Test for unusual characters */
			if (!quiet && ! alphabetic(forenames))
				fprintf(err, "Warning: row %d, forename(s) = %s contains unusual characters\n",
					n, forenames), nwarn++;

			if (forenames[0] == (char)NULL)
			{
				if (! quiet && name_err[0] != 'x')
					fprintf(err, "Warning: row %d, forename(s) missing\n",
						n), nwarn++;
				strcpy(forenames, "-");
				strcpy(name_err, "x");
			}
			else
				inits(forenames);

			if (name_err[0] != 'x' && name_err[0] != '-')
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, name flag = %s invalid value\n",
						n, name_err), ++nwarn;
				strcpy(name_err, "x");
			}

			/****************/
			/* detail group */
			/****************/

			if (detail_err[0] == (char)NULL)
				strcpy(detail_err, "-");
			else
				detail_err[0] = tolower(detail_err[0]);

			if (year > 1841)
			{
				/* 4.9.1 Test for unusual characters */
				if (!quiet && ! alphabetic(rel))
					fprintf(err, "Warning: row %d, relationship = %s contains unusual characters\n",
						n, rel), nwarn++;

				if (rel[0] == (char)NULL)
				{
					if (! quiet && detail_err[0] != 'x')
						fprintf(err, "Warning: row %d, relationship missing\n",
							n), nwarn++;
					strcpy(rel, "-");
					strcpy(detail_err, "x");
				}
				else
				{
					inits(rel);

					/* 4.10.0 check that we have a valid relationship code */
					if (rcheck && strcmp(rel, "-") != 0 && cfind(rcode, rel, 6) < 0)
					{
						if (! quiet)
							fprintf(err, "Warning: row %d, unrecognised relationship code %s\n",
								n, rel), nwarn++;
						/* 4.10.8 don't flag these, creates too much green!
						strcpy(detail_err, "x"); */
					}
				}

				if (! quiet && strcmp(rel, "Head") == 0 && ! new_household)
						fprintf(err,
	"Warning: row %d, Head of household is not the first entry in the schedule\n",
							n), nwarn++;

				if (cond[0] == (char)NULL)
					/* silently substitute because it is so often missing */
					strcpy(cond, "-");
				else
					cond[0] = toupper(cond[0]);

				if (cond[0] != 'M' && cond[0] != 'S' && cond[0] != 'U'
					&& cond[0] != 'W' && cond[0] != '-')
				{
					if (! quiet)
						fprintf(err,
							"Warning: row %d, condition = %s invalid value\n",
							n, cond), nwarn++;
					strcpy(detail_err, "x");
				}
			}
			else
			{
				if (verbose && rel[0] != (char)NULL && strcmp(rel, "-") != 0)
					fprintf(err, "Info: row %d, relationship ignored\n", n);
				strcpy(rel, "-");
				if (verbose && cond[0] != (char)NULL && cond[0] != '-')
					fprintf(err, "Info: row %d, condition ignored\n", n);
				strcpy(cond, "-");
			}

			if (sex[0] == (char)NULL)
			{
				if (! quiet)
					fprintf(err, "Warning: row %d, sex marker missing\n", n), nwarn++;
				strcpy(sex, "-");
				strcpy(detail_err, "x");
			}
			else
				sex[0] = toupper(sex[0]);

			if (sex[0] != 'M' && sex[0] != 'F' && sex[0] != '-')
			{
				/* 4.10.6 Make this an Error */
				fprintf(err, "Error: row %d, sex = %s invalid value\n",
						n, sex), nerr++;
				strcpy(sex, "-");
			}

			/* extract the units part of the age field */
			age_u = part2(age);

			if (age[0] == (char)NULL)
			{
				if (! quiet && detail_err[0] != 'x')
					fprintf(err, "Warning: row %d, age missing\n", n), nwarn++;
				strcpy(age, "999");
				strcpy(detail_err, "x");
			}

			if (! numeric(age))
			{
				fprintf(err, "Error: row %d, age = %s not numeric\n",
						n, age), nerr++;
			/* 4.10.5 Put in a valid age */
				strcpy(age, "999");
				age_u = 'y';
			}
			nage = atoi(age);

			if (age_u == ' ')
				age_u = 'y';
			else
				age_u = tolower(age_u);

			/* 4.5.1 make sure 999 is reflected in error flag */
			if (nage == 999)
			{
				age_u = 'y';
				strcpy(detail_err, "x");
			}
				
			/* 4.7.5 check for zero age */
			if (nage == 0)
			{
				if (! quiet)
					fprintf(err, "Warning: row %d, zero age\n", n), nwarn++;
				strcpy(detail_err, "x");
			}				

			switch (age_u)
			{
				case 'y':
					if (nage != 999 && nage > 120)
					{
						if (! quiet)
							fprintf(err, "Warning: row %d, age more than 120y\n",
								n), nwarn++;
						strcpy(detail_err, "x");
					}
					break;
				case 'm':
					if (nage > 24)
					{
						if (! quiet)
							fprintf(err, "Warning: row %d, age more than 24m\n",
								n), nwarn++;
						strcpy(detail_err, "x");
					}
					break;
				case 'w':
					if (nage > 20)
					{
						if (! quiet)
							fprintf(err, "Warning: row %d, age more than 20w\n",
								n), nwarn++;
						strcpy(detail_err, "x");
					}
					break;
				case 'd':
					if (nage > 30)
					{
						if (! quiet)
							fprintf(err, "Warning: row %d, age more than 30d\n",
								n), nwarn++;
						strcpy(detail_err, "x");
					}
					break;
				default:
					if (! quiet)
						fprintf(err,
							"Warning: row %d, age units = %c invalid value\n",
							n, age_u), nwarn++;
					strcpy(detail_err, "x");
			}

			if (year > 1841 && (cond[0] == 'M' || cond[0] == 'W')
				&& (nage < 14 || age_u != 'y'))
			{
				if (! quiet)
					fprintf(err, "Warning: row %d, married/widowed person age under 14y\n",
						n), nwarn++;
				strcpy(detail_err, "x");
			}

			if (detail_err[0] != 'x' && detail_err[0] != '-')
			{
				if (! quiet)
					fprintf(err,
					"Warning: row %d, detail flag = %s invalid value\n",
						n, detail_err), ++nwarn;
				strcpy(detail_err, "x");
			}

			/********************/
			/* occupation group */
			/********************/

			if (occ_err[0] == (char)NULL)
				strcpy(occ_err, "-");
			else
				occ_err[0] = tolower(occ_err[0]);

			/* no default for occupation - blank is acceptable */
			inits(occ);

			/* consolidate the employment flag into the occupation field */
			if (brief)
			{
				if (year > 1881)
				{
					if (emp[0] == (char)NULL)
						strcpy(emp, "-");
					else
						emp[0] = toupper(emp[0]);

					if (emp[0] != ' ' && emp[0] != '-')
					{
						/* note that 24 + 7 = 31 */
						if (strlen(occ) > 24)
						{
							fprintf(err,
				"Error: row %d, no room for employed status in occupation field\n",
								n), nerr++;
							strcpy(occ_err, "x");
						}
						else if (emp[0] == 'E')
							strcat(occ, "(Em'ee)");
						else if (emp[0] == 'R')
							strcat(occ, "(Em'er)");
						else if (emp[0] == 'N')
							strcat(occ, "(Notem)");
						else
						{
							if (! quiet)
								fprintf(err,
						"Warning: row %d, employed status = %s invalid value\n",
									n, emp), nwarn++;
							strcpy(occ_err, "x");
						}
						/* If necessary, chop off the trailing ")"
						This is why the variable is 32 characters long */
						occ[30] = (char)NULL;
					}
				}
				else if (verbose && emp[0] != (char)NULL)
					fprintf(err, "Info: row %d, employed status ignored\n", n);
			}

			if (occ_err[0] != 'x' && occ_err[0] != '-')
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, occupation flag = %s invalid value\n",
						n, occ_err), ++nwarn;
				strcpy(occ_err, "x");
			}

			/********************/
			/* birthplace group */
			/********************/

			if (place_err[0] == (char)NULL)
				strcpy(place_err, "-");
			else
				place_err[0] = tolower(place_err[0]);

			if (chap[0] == (char)NULL)
			{
				if (! quiet)
					fprintf(err, "Warning: row %d, county code missing\n",
						n), nwarn++;
				strcpy(chap, "UNK");
				strcpy(place_err, "x");
			}
			else
				(void)capitals(chap);

			if (strlen(chap) != 3)
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, county code = %s invalid length\n",
						n, chap), nwarn++;
				strcpy(place_err, "x");
			}

			/* 4.6.0 check that we have a valid chapman code */
			else if (ccheck && cfind(ccode, chap, 3) < 0)
			{
				if (! quiet)
					fprintf(err, "Warning: row %d, unrecognised county code %s\n",
						n, chap), nwarn++;
				strcpy(place_err, "x");
			}

			if (year > 1841)
			{
				/* 4.9.1 Test for unusual characters */
				if (!quiet && ! alphabetic(place))
					fprintf(err, "Warning: row %d, birth place = %s contains unusual characters\n",
						n, place), nwarn++;

				if (place[0] == (char)NULL)
				{
					if (verbose && place_err[0] != 'x')
					{
						fprintf(err, "Info: row %d, birth place missing\n", n);
						strcpy(place_err, "x");
					}
					strcpy(place, "-");
				}
				else
					inits(place);
			}
			else
			{
				if (verbose && place[0] != (char)NULL && strcmp(place, "-") != 0)
					fprintf(err, "Info: row %d, birth place ignored\n", n);
				strcpy(place, "-");
			}

			if (place_err[0] != 'x' && place_err[0] != '-')
			{
				if (! quiet)
					fprintf(err,
						"Warning: row %d, birth place flag = %s invalid value\n",
						n, place_err), ++nwarn;
				strcpy(place_err, "x");
			}

			if (rlen > 276)
			{
				if (altchap[0] != (char)NULL)
				{
					(void)capitals(altchap);
					if (strlen(altchap) != 3)
					{
						if (! quiet)
							fprintf(err,
								"Warning: row %d, alternate county code = %s invalid length\n",
								n, altchap), nwarn++;
						/* 4.6.1 */
						strcpy(place_err, "x");
					}

					/* 4.6.0 check that we have a valid chapman code */
					else if (ccheck && cfind(ccode, altchap, 3) < 0)
					{
						if (! quiet)
							fprintf(err, "Warning: row %d, unrecognised alternate county code %s\n",
								n, altchap), nwarn++;
						strcpy(place_err, "x");
					}
				}


				if (year > 1841)
				{
				/* 4.9.1 Test for unusual characters */
					if (!quiet && ! alphabetic(altplace))
						fprintf(err, "Warning: row %d, alternate birth place = %s contains unusual characters\n",
							n, altplace), nwarn++;

					if (altplace[0] != (char)NULL)
						inits(altplace);
				}
				else
				{
					if (verbose && altplace[0] != (char)NULL)
						fprintf(err, "Info: row %d, alternate birth place ignored\n", n);
					altplace[0] = (char)NULL;
				}
			}

			/**********************/
			/* miscelaneous group */
			/**********************/

			if (year > 1841)
			{
				/* 4.9.1 Test for unusual characters */
				if (!quiet && ! alphabetic(dis))
					fprintf(err, "Warning: row %d, disability = %s contains unusual characters\n",
						n, dis), nwarn++;

				if (dis[0] == (char)NULL)
					strcpy(dis, " ");
				else
					inits(dis);
			}
			else
			{
				if (verbose && dis[0] != (char)NULL)
					fprintf(err, "Info: row %d, disability ignored\n", n);
				dis[0] = (char)NULL;
			}

			/* 4.8.6 Correctly deal with the Language flag */
			if (wls)
			{
				/* The valid values for the welsh flag are "W", "E" or "B" */
				if (welsh[0] != 'W' && welsh[0] != 'E' && welsh[0] != 'B'
					&& welsh[0] != '-')
				{
					if (! quiet)
						fprintf(err,
							"Warning: row %d, language = <%s> invalid value\n",
							n, welsh), nwarn++;
					strcpy(welsh, "x");
				}
			}
			else if (sct && year >= 1881)
			{
				/* 4.8.8 Gaelic. The valid values for the flag are "G" or "B"
					The actual question was "G" or "G & E"? */
				if (welsh[0] != 'G' && welsh[0] != 'B' && welsh[0] != '-')
				{
					if (! quiet)
						fprintf(err,
							"Warning: row %d, language = <%s> invalid value\n",
							n, welsh), nwarn++;
					strcpy(welsh, "x");
				}
			}
			else
			{
				if (verbose && welsh[0] != (char)NULL)
					fprintf(err, "Info: row %d, Language ignored\n", n);
				strcpy(welsh, " ");
			}
		}

		/* Save some of the fields now (before we pad them out)
			so they are available for the next record */
		/* 4.9.0 change regno fields to mean book number */
		strcpy(p_book, book);
		strcpy(p_parish, parish);
		if (rlen > 256)
			strcpy(p_parishe, parishe);
		strcpy(p_ed, o_ed);
		strcpy(p_folio, o_folio);
		strcpy(p_page, page);
		strcpy(p_sched, o_sched);
		strcpy(p_house_num, o_house_num);
		strcpy(p_house_name, house_name);
		strcpy(p_buvn, buvn);

		/* 4.4.8 The database does not allow more than three digits so utilise
		the alpha suffix to allow for the occasional overflow */

		if (atoi(sched) > 999)
		{
			if (atoi(sched) <= 1999)
			{
				sched_a = '!';
				sprintf(sched, "%d", atoi(sched) - 1000);
			}
			else
			{
				fprintf(err, "Error: row %d, schedule > 1999, \n",
					n), nerr++;
				strcpy(sched, "999");
			}
		}

		/*************************/
		/* output the new record */
		/*************************/

		fputc(del[0], out);

		/* 4.9.0 this (padding with nulls) is no longer needed
		fputs(book, out);
		for (i = 4 - strlen(book); i > 0; --i)
			fputc((char)NULL, out);
*/
		fprintf(out, "%s", pad(book, 4));
		/* 4.10.4 Recover the original value of book because it may not
			be read in again next time around */
		strcpy(book, p_book);

		fprintf(out, "%0.6d%0.4d%s%s%c%s%c%s%s%c%s%c%s%c",
			atoi(household), atoi(household_mem), pad(parish, 20), pad(ed, 3), ed_a,
			pad(folio, 4), folio_a, pad(page, 4),
			pad(sched, 3), sched_a, pad(house_num, 4),
			house_num_a, pad(house_name, 30), buvn[0]);
		fprintf(out, "%s%s%c%s%c%c%s%c%c",
			pad(surname, 24), pad(forenames, 24), name_err[0], pad(rel, 6),
			cond[0], sex[0], pad(age, 3), age_u, detail_err[0]);
		fprintf(out, "%s%c%s%s%c%s%c%s",
			pad(occ, 30), occ_err[0], pad(chap, 3), pad(place, 20), place_err[0],
			pad(dis, 6), welsh[0], pad(notes, 44));
		if (rlen > 256)
			fprintf(out, "%s", pad(parishe, 20));
		if (rlen > 276)
			fprintf(out, "%s%s", pad(altchap, 3), pad(altplace, 20));

	/* ******************* */
	/* END OF THE BIG LOOP */
	/* ******************* */
	}

	/* Munge the record count so that it makes sense */
	if (brief)
	{
		n -= offset;
		maxhh = atoi(household);
	}

	fprintf(err, "\nOutput %d records, highest household number = %d\n",
		n, maxhh);
	fprintf(err, "\nOutput file is %s\n", output);

	fprintf(stderr, "\nOutput %d records, highest household number = %d\n",
		n, maxhh);
	fprintf(stderr, "\nOutput file is %s\n", output);
	fprintf(stderr, "\n %d errors", nerr);
	if (! quiet)
		fprintf(stderr, ", %d warnings", nwarn);
	fprintf(stderr, ". see %s\n", error);

	fclose(err);
	fclose(out);
	fclose(in);

	if (nerr > 0)
		return 8;
	if (nwarn > 0)
		return 4;

	return 0;
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash (unix) or a slash (DOS) */
int options(int argc,char *argv[])
{
	int i,firstnonoption=0;
	extern int brief, rlen, verbose, quiet, rcheck;

	brief = TRUE;
	rlen = 256;
	verbose = FALSE;
	quiet = FALSE;

	for (i=1; i < argc;i++)
	{
		if (
#ifndef UNIX
			argv[i][0] == '/' ||
#endif
			argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
				case 'a':
				case 'A':
					brief = FALSE;
					break;
				case 'b':	/* for backwards compatability */
				case 'B':
					brief = TRUE;
					break;
				case 'c':
				case 'C':
					rlen = 276;
					break;
				case 'd':
				case 'D':
					rlen = 299;
					break;
				case 'h':
				case 'H':
				case '?':
					usage();
					return 0;
				/* 4.10.0 */
				case 'r':
				case 'R':
					rcheck = TRUE;
					break;
				case 'v':
				case 'V':
					verbose = TRUE;
					quiet = FALSE;
					break;
				case 'q':
				case 'Q':
					verbose = FALSE;
					quiet = TRUE;
					break;
				default:
					fprintf(stderr,"Unknown option %s\n",argv[i]);
					break;
			}
		}
		else
		{
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

/* extract a value from a CSV field allowing for quotes etc. */
char * extract(char * buff, int * off, int len)
{
	/* note: common output buffer, do not call twice in the same expression */
	extern char hold[];
	int i;
	char end;
	/* extract_err and extract_content are reset externally from this routine so
		they are cumulative */
	extern int n, extract_err, extract_content, quiet, nerr;
	extern FILE * err;
	end = ',';
	i = 0;

	/* 4.10.7 strip leading spaces before all else */
	for ( ; buff[*off] == ' '; (*off)++)
		;

	if (buff[*off] == '"')
	{
		(*off)++;
		end = '"';
	}

	while (TRUE)
	{
		while (buff[*off] != end && buff[*off] != (char)NULL)
		{
			hold[i] = buff[(*off)++];
		/* drop leading spaces - also supress leading escape character
		This is done because some spreadsheets bleat about fields that start
		with a dash */
			if (i == 0 && (hold[i] == ' ' || hold[i] == '\\'))
				continue;
		/* 4.8.10 supress consecutive spaces (most often in the notes field) */
			if (i != 0 && hold[i] == ' ' && hold[i - 1] == ' ')
				continue;
		/* 4.11.0 Fundamental valid character test */
			if (! isvalid((unsigned char)hold[i]))
			{
				if (! brief || n > 3)
					fprintf(err, "Error: row %d, field contains an invalid character <%c>\n",
						n, hold[i]), nerr++;
				extract_err = TRUE;
			}
			i++;
		}
		if (buff[*off] == (char)NULL)
			break;

		(*off)++;

		if (end == '"')
		{
			/* Two quotes in a row pass one through */
			if (buff[*off] == '"')
			{
				hold[i++] = '"';
				(*off)++;
			}
			else
				/* now look for the field break */
				end = ',';
		}
		else
			break;
	}
	if (end != ',')
	{
		fprintf(err, "Error: row %d, ends with incomplete field\n", n), nerr++;
		extract_err = TRUE;
	}

	/* remove trailing blanks */
	while (i > 0 && hold[i - 1] == ' ')
		--i;

	hold[i] = (char)NULL;

	if (i > len)
	{
		/* 4.8.7 increase severity to Error */
		/* 4.10.1 say how big it should have been */
		fprintf(err, "Error: row %d, field too long <%s>, truncated by %d to %d characters in the output file\n", n, hold, i - len, len), nerr++;
		hold[len] = (char)NULL;
		extract_err = TRUE;
	}

	/* convert the alternate missing item flag (some spreadsheets don't like -) */
	if (strcmp(hold, "/") == 0)
		strcpy(hold, "-");

	/* say if we found anything */
	if (*hold != (char)NULL)
		extract_content = TRUE;

	return hold;
}

/* 4.5.0 remove trailing question marks etc. */
void noquery(char * field, int f, char * field_err)
{
	int i;
	extern FILE * err;

	i = strlen(field);

	/* remove trailing question marks */
	if (i > 0 && field[i - 1] == '?')
	{
		--i;
		if (f)
		{
			strcpy(field_err, "x");
			if (! quiet)
				fprintf(err, "Warning: row %d, trailing \"?\" removed and flag set.\n", n), nwarn++;
		}
		else
			if (! quiet)
				fprintf(err, "Warning: row %d, trailing \"?\" removed.\n", n), nwarn++;
		/* remove any more while we are at it */
		while (field[i - 1] == '?')
			--i;
		/* remove any more trailing blanks */
		while (field[i - 1] == ' ')
			--i;
	}

	field[i] = (char)NULL;

	/* 4.6.5 further check to see if there are any ? marks at all */
	if (f && field_err[0] != 'x' && strchr(field, '?') != (char *)NULL)
	{
		strcpy(field_err, "x");
		if (! quiet)
			fprintf(err, "Warning: row %d, \"?\" detected, flag set.\n", n), nwarn++;

	}

	return;
}

/* 4.6.0 routine to check if the Chapman code is recognised
	4.10.0 actually works for any list of strings */
int cfind(char * ccode, char * code, int n)
{
	char * cp;
	for (cp = ccode; *cp != (char)NULL && strcmp(cp, code) != 0; cp += n + 1)
		;
	if (*cp == (char)NULL)
		return -1;
	return 1;
}


/* fill out a string with spaces to the maximum length */
char * pad(char * string, int len)
{
	char * p;
	int i, l;
	l = strlen(string);
	p = string + l;
	for (i = len - l; i > 0; --i)
		*(p++) = ' ';
	*p = (char)NULL;

	return string;
}

/* test to see if a value is numeric */
int numeric(char * string)
{
	char * p = string;
	while (*p != (char)NULL)
		if (! isdigit(*p++))
			return FALSE;
	return TRUE;
}

/* extract the trailing alphabetic character from a field (if any) */
char part2(char * string)
{
	int i;
	char c;
	i = strlen(string) - 1;
	if (i < 0)
	return ' ';
	if (! isdigit(string[i]))
		c = string[i];
	else if (i > 0 && string[i - 1] == '#')
	/* special case for coded sections (institutions & shipping etc.) */
		c = string[i--];
	else
		return ' ';

	/* remove trailing blanks */
	while (string[i - 1] == ' ')
		--i;
	string[i] = (char)NULL;

	return c;
}

/* capitalise a string (8 bit version - allows for accented characters) */
char * capitals(char * string)
{
	int i = 0;

	while (string[i] != (char)NULL)
		string[i++] = (char)toupper8((unsigned char)string[i]);

	return string;
}

/* a routine that capitalises the first letter of each word (8 bit version) */
void inits(char * string)
{
	int i = 0;
	char last = ' ';

	while (string[i] != (char)NULL)
	{
		if (last == ' ' || last == '(' || last == '-' || last == '.' || last == ',')
		{
			if (string[i] == '\"' || string[i] == '\'')
			{
				i++;
				continue;
			}
			string[i] = (char)toupper8((unsigned char)string[i]);
		}
		else
			string[i] = (char)tolower8((unsigned char)string[i]);
		last = string[i++];
	}

	return;
}

void splitpath(const char * string, char * path, char * file, char * ext)
{
	char * cp;

	strcpy(path, string);
	*file = (char)NULL;
	*ext = (char)NULL;

#ifdef UNIX
	if ((cp = strrchr(path, '/')) != (char *)NULL)
#else
	if ((cp = strrchr(path, '\\')) != (char *)NULL)
#endif
	{
		*cp = (char)NULL;
		strcpy(file, cp + 1);
	}
	else
	{
		strcpy(file, path);
		*path = (char)NULL;
	}

	if ((cp = strchr(file, '.')) != (char *)NULL)
	{
		*cp = (char)NULL;
		strcpy(ext, cp + 1);
	}

	return;
}

char * makepath(const char * path, const char * file, const char * ext)
{
	extern char hold[];

	*hold = (char)NULL;
	if (*path != (char)NULL)
	{
		strcpy(hold, path);
#ifdef UNIX
		strcat(hold, "/");
#else
		strcat(hold, "\\");
#endif
	}
	strcat(hold, file);
	strcat(hold, ".");
	strcat(hold, ext);

	return hold;
}

char * fgetm(char * line, int maxline, FILE * fp)
{
	/* fgetm does the same as fgets but allows for PC files
		where the lines end in \r\n, MAC files where the lines
		end in \r and UNIX files where the lines end in \n.
		This routine will insert a standard \n on the end of
		each line. */

	register int c;
	register char *cs;

	cs = line;
	while (--maxline > 0 && (c = getc(fp)) != EOF)
		if ((*cs++ = c) == '\n' || c == '\r')
			break;

	if (c == '\r')
	{
		*(cs - 1) = '\n';
		c = getc(fp);
		if (c != '\n' && c != EOF)
			ungetc(c, fp);
	}
	*cs = (char)NULL;
	return (c == EOF && cs == line) ? (char *)NULL : line;
}

/* return code is passed back as return code of main */
int decode_census(char * fname)
{
	extern int year, sct, ipart;
	extern FILE * err;
	/* 4.7.3 extend all path and file names to 128 (XP likes long paths) */
	char Fname[128], class[15], piece[10], padpiece[10], part[3];
	int i, con;

	con = sct = FALSE;

	/* take a copy in capitals to make it easier */
	strcpy(Fname, fname);
	(void)capitals(Fname);
	strcpy(part, "00");

	/* This could be more compact, but is like this for clarity */
	if (strncmp(Fname, "HO1", 3) == 0) /* Cornwall & Original 1841 */
	{
		year = 1841;
		strcpy(fname, "ho1");
		/* this is a fiddle to allow CON to use the new part numbering scheme
			and also caters for the old HO107 scheme */
		if (Fname[4] > '6')
		{
			con = TRUE;
			/* The problem is that we had already divided the pieces into more
			than 9 parts so we use the next digit as well. The sequence
			is 17 ... 97, 18 ... 98, 19 ... 99  (08 and 09 are not used )
			so this allows for 27 parts. It works because the piece numbers
			are only three digits. 07 caters for the old HO107 pattern */
			strncpy(part, Fname + 3, 2);
			ipart = (part[1] - '7') * 10 + part[0] - '0';
			if (ipart == 10 || ipart == 20)
			{
				fprintf(err,
					"Fatal: Part number in file name (%s) is not valid\n",
					part);
				fprintf(stderr,
					"Fatal: Part number in file name is not valid\n");
				return 12;
			}
			/* fill in the gaps not used */
			if (ipart > 20) --ipart;
			if (ipart > 10) --ipart;
			strcpy(piece, Fname + 5);

			/* 4.8.0 revert original HO107 to be not CON */
			if (ipart == 0)
			{
				con = FALSE;
				strcpy(fname, "ho107");
				strcpy(part, "00");
			}
		}
		else /* 4.8.0 the next allows for HO10nnnn which was allowed for a
			short while by accident - I don't think it was ever used */
		{	strcat(fname, "0");
			strcpy(piece, Fname + 4);
		}
		strcpy(class, "HO107");
	}
	/* v 4.8.0 allow proper 1841 with 26 sub parts */
	else if (strncmp(Fname, "HO4", 3) == 0) /* England/Wales 1841 */
	{
		year = 1841;
		strcpy(fname, "ho41");
		/* Allow 26 sub parts */
		if (Fname[3] >= 'A' && Fname[3] <= 'Z')
		{
			sprintf(part, "%0.2d", Fname[3] - 'A' + 1);
			fname[3] = 'a' + Fname[3] - 'A';
		}
		strcpy(piece, Fname + 4);
		strcpy(class, "HO107");
	}
	else if (strncmp(Fname, "HS4", 3) == 0) /* Scotland 1841 */
	{
		year = 1841;
		sct = TRUE;
		strcpy(fname, "hs4");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1841"); /* bogus */
	}
	else if (strncmp(Fname, "HO5", 3) == 0) /* England/Wales 1851 */
	{
		year = 1851;
		strcpy(fname, "ho51");
		/* Allow 26 sub parts */
		if (Fname[3] >= 'A' && Fname[3] <= 'Z')
		{
			sprintf(part, "%0.2d", Fname[3] - 'A' + 1);
			fname[3] = 'a' + Fname[3] - 'A';
		}
		strcpy(piece, Fname + 4);
		strcpy(class, "HO107");
	}
	else if (strncmp(Fname, "HS5", 3) == 0) /* Scotland 1851 */
	{
		year = 1851;
		sct = TRUE;
		strcpy(fname, "hs5");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1851"); /* bogus */
	}
	else if (strncmp(Fname, "RG09", 4) == 0) /* England/Wales 1861 */
	{
		year = 1861;
		strcpy(fname, "rg09");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG9");
	}
	else if (strncmp(Fname, "RG9", 3) == 0) /* just in case */
	{
		year = 1861;
		strcpy(fname, "rg09");
		strcpy(piece, Fname + 3);
		strcpy(class, "RG9");
	}
	else if (strncmp(Fname, "RS6", 3) == 0) /* Scotland 1861 */
	{
		year = 1861;
		sct = TRUE;
		strcpy(fname, "rs6");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1861"); /* bogus */
	}
	else if (strncmp(Fname, "RG10", 4) == 0) /* England/Wales 1871 */
	{
		year = 1871;
		strcpy(fname, "rg10");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG10");
	}
	else if (strncmp(Fname, "RS7", 3) == 0) /* Scotland 1871 */
	{
		year = 1871;
		sct = TRUE;
		strcpy(fname, "rs7");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1871"); /* bogus */
	}
	else if (strncmp(Fname, "RG11", 4) == 0) /* England/Wales 1881 */
	{
		year = 1881;
		strcpy(fname, "rg11");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG11");
	}
	else if (strncmp(Fname, "RS8", 3) == 0) /* Scotland 1881 */
	{
		year = 1881;
		sct = TRUE;
		strcpy(fname, "rs8");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1881"); /* bogus */
	}
	else if (strncmp(Fname, "RG12", 4) == 0) /* England/Wales 1891 */
	{
		year = 1891;
		strcpy(fname, "rg12");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG12");
	}
	else if (strncmp(Fname, "RS9", 3) == 0) /* Scotland 1891 */
	{
		year = 1891;
		sct = TRUE;
		strcpy(fname, "rs9");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1891"); /* bogus */
	}
	else if (strncmp(Fname, "RG13", 4) == 0) /* England/Wales 1901 */
	{
		year = 1901;
		strcpy(fname, "rg13");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG13");
	}
	else if (strncmp(Fname, "RS0", 3) == 0) /* Scotland 1901 */
	{
		year = 1901;
		sct = TRUE;
		strcpy(fname, "rs0");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1901"); /* bogus */
	}
	else
	{
		fprintf(err, "Fatal: Census year cannot be determined from file name (%s)\n", fname);
		fprintf(stderr, "Fatal: Census year cannot be determined from file name\n");
		return 12;
	}

	/* check that part number is numeric */
	if (! numeric(part))
	{
		fprintf(err, "Fatal: Part number in file name (%s) is not numeric\n",
			part);
		fprintf(stderr, "Fatal: Part number in file name is not numeric\n");
		return 12;
	}

	/* this is already done for CON 1841 which uses a different algorithm */
	if (! con)
		ipart = atoi(part);

	if (sct)
	{
		if (ipart < 10)
		{
			fprintf(err, "Fatal: Part number in file name (%s) is not valid\n",
				part);
			fprintf(stderr, "Fatal: Part number in file name is not valid\n");
			return 12;
		}
		ipart -= 10;
	}

	/* check that piece number is numeric */
	if (! numeric(piece))
	{
		fprintf(err, "Fatal: Piece number in file name (%s) is not numeric\n",
			piece);
		fprintf(stderr, "Fatal: Piece number in file name is not numeric\n");
		return 12;
	}

	/* 4.8.0 this section all juggled around a bit */

	if (con || sct)
		strcat(fname, part);

	/* tidy up piece number */
	i = atoi(piece);
	sprintf(piece, "%d", i);

	if (strlen(fname) == 5)
		sprintf(padpiece, "%0.3d", i);
	else
		sprintf(padpiece, "%0.4d", i);

	/* finish off file names */
	strcat(fname, padpiece);
	return 0;
}

/* 8 bit version of tolower to allow for accented characters ISO8859-1 */
unsigned char tolower8(unsigned char c)
{
	/*         A  to       Z           Agrave   to  THORN   not multiply */
	if ((c >= 'A' && c <= 'Z') || (c >= 0xC0 && c <= 0xDE && c != 0xD7))
		return c + 0x20;
	return c;
}

/* 8 bit version of toupper to allow for accented characters ISO8859-1 */
unsigned char toupper8(unsigned char c)
{
	/*         a  to       z           agrave   to  thorn    not divide */
	if ((c >= 'a' && c <= 'z') || (c >= 0xE0 && c <= 0xFE && c != 0xF7))
		return c - 0x20;
	return c;
/* if just testing lower we would need to include DF (szlig) and FF (yuml) */
}

/* 4.9.1 Routine to check for extended alphabetic - suitable for names and addresses */
int alphabetic(char * string)
{
	char * p = string;
	while (*p != (char)NULL)
		if (! isalpha(*p) && ! isusual(*p))
			return FALSE;
		else p++;
	return TRUE;
}

int isusual(char c)
{	/* 4.9.2 add & to the list (common in disability column)
	   4.10.2 add comma */
	if (c == ' ' || c == '.' || c == '-' || c == '(' || c == ')' || c == '\'' || c == '&' || c == ',')
		return TRUE;
	return FALSE;
}

/* 4.11.0 */
int isvalid(unsigned char c)
{	/* The valid character set is a subset of ISO8859-1 */
	  /* Printable 7 bit Ascii */
	if ((c >= ' ' && c <= 0x7F) ||
		/* agrave to yuml      not divide */
		(c >= 0xE0 && c <= 0xFF && c != 0xF7) ||
		/* Agrave to szlig   not multiply */
		(c >= 0xC0 && c <= 0xDF && c != 0xD7))
/* Strictly eth (DO & FO) and thorn (DE & FE) should also be omitted */
		return TRUE;
	return FALSE;
}
