/* dat2csv */
#define VERSION "4.7.4"

/* Uncomment the next line or use -DUNIX to compile on a unix system */
/* #define UNIX */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

void usage(void)
{
	fprintf(stderr,"Usage (v%s): dat2csv [-a|-w] [-q|-v] [-z] [-h|-?] <filename.ext>\n", VERSION);
	fprintf(stderr,"\ta=ALL, w=WEB, q=QUIET, v=VERBOSE, z=STRIP, h|?=HELP\n");
}

/* Copyright (c) 2002-05 Rick Parsons.

	This software is published Open Source Freeware. You are free to use the
	programs either in full or part without charge. I would ask however that
	this notice is included with any extract or onward distribution.

	No charge may be made for the use of this code, however there is no
	objection to including modest extracts (with acknowledgement) in otherwise
	commercial software.

	No warranty is given or implied by the use of the software which you do at
	your own risk. No compensation can be considered regarding damage to data,
	computers or any thing else arising as a result of using these programs.
*/

#define TRUE 1
#define FALSE 0

#define b_del 0
/* 4.7.0 change reg field to mean book number */
#define b_book 1
#define b_household 5
#define b_household_mem 11
#define b_parish 15
#define b_ed 35
#define b_ed_a 38
#define b_folio 39
#define b_folio_a 43
#define b_page 44
#define b_sched 48
#define b_sched_a 51
#define b_house_num 52
#define b_house_num_a 56
#define b_house_name 57
#define b_buvn 87
#define b_surname 88
#define b_forenames 112
#define b_name_err 136
#define b_rel 137
#define b_cond 143
#define b_sex 144
#define b_age 145
#define b_age_unit 148
#define b_detail_err 149
#define b_occ 150
#define b_occ_err 180
#define b_chap 181
#define b_place 184
#define b_place_err 204
#define b_dis 205
#define b_welsh 211
#define b_notes 212
#define b_parishe 256
#define b_altchap 276
#define b_altplace 279

int options(int, char **);
void strip(char *);
char * prepare(char *, char);
char * squash(char, char);
/* 4.6.10 to remove unexpected commas */
char * sprep(char *, char);
int cfind(char * *, char *);
int webhead(FILE *, char *, char *, char *, char *);
void webfoot(FILE *, char *, char *, char *, char *);
void escape(char, FILE *, char *, char *, char *, char *);
void splitpath(const char *, char *, char *, char *);
char * makepath(const char *, const char *, const char *);
/* 4.5.1 change to unsigned */
int vote(unsigned char *);
int maxvote(unsigned char *, int);
int numeric(char *);
char * capitals(char *);
int decode_census(char *, char *, char *);
char decode_emp(char *);
/* new for 4.5.1 */
int isupper8(unsigned char);
unsigned char tolower8(unsigned char);

static char hold[50];
static char hold2[10];
static char date[40];


/* The vote mask definitions.
	x = an error field (x or -)
	n = numeric (0-9)
	A = upper case alpha (A-Z, upper case accented or -)
*/
#define MAXVOTE 10

static struct mask
{	int p;
	char t;
} vmasks[MAXVOTE] =
{	{b_household, 'n'},
	{b_ed, 'n'},
	{b_folio, 'n'},
	{b_page, 'n'},
	{b_sched, 'n'},
	{b_name_err, 'x'},
	{b_detail_err, 'x'},
	{b_occ_err, 'x'},
	{b_place_err, 'x'},
	{b_surname, 'A'}
};

#define LEEWAY b_parish

#define ALL 2
#define WEB 4
#define BRIEF 16
static int year, mode, verbose, quiet, nerr, nwarn, con, dstrip, ipart, sct;
static FILE * tmp, * err;
/* 4.5.4 ppath = program path - where this binary is located */
static char ppath[128];

int _main(int argc, char *argv[])
{
	long foffset;
	FILE * in;
	FILE * out;
	extern FILE * err;
	char * cp;
	unsigned char * ucp;
	/* 4.7.0 Add a marker for good book numbers */
	int i, j, n, p, v, t, rlen, firstp = TRUE, ret, wls = FALSE, gbook = FALSE;
	char c;
	extern int mode, year, verbose, quiet, nerr, nwarn, con, dstrip, sct;
	extern char hold[];
	extern char date[];

	/* 4.5.3 extend all path and file names to 128 (XP likes long paths) */
	char input[128];
	char output[128];
	char error[128];
	char temp[128];
	char fname[128], Fname[15];
	char webname[15], proref[25];
	char path[128];
	char ext[128];
	extern char ppath[];

/* 4.5.1 unsigned to allow for 8 bit accented characters in data */

 	unsigned char realbuf[650]; /* at least twice max data record + LEEWAY (613) */

/* 4.5.1 this points to part of realbuf and should also be unsigned but so many of
 	the routines want signed that it becomes too hard. Where it is important
 	there is a cast */

	char * buf;

	char work[50];
	time_t tod;

	/* 4.7.0 change reg field to mean book number */
	char book[5];
	char household[7], o_household[7];
	char household_mem[5];
	char parish[21], o_parish[21];
	char ed[5], o_ed[5], o_ed_a = 'X'; /* allow room for # as well */
	char folio[5], o_folio[5], o_folio_a = 'X';
	char page[5], o_page[5];
	char sched[5]; /* 4.4.8 allow room for 4 digits */
	char house_num[5];
	char house_name[31];
	char surname[25];
	char forenames[25];
	char rel[7];
	char age[4];
	char occ[31];
	char chap[4];
	char place[21];
	char dis[7];
	char notes[45];
	char parishe[21], o_parishe[21];
	char emp[9];
	char altchap[4];
	char altplace[21];

	char * * ccode;
	char * * county;
	char * block;
	/*
		(1) Barracks & Military Qtrs
		(2) HM Ships at Home
		(3) Workhouses (inc pauper schools)
		(4) Hospitals
		(5) Lunatic Asylums
		(6) Prisons
		(7) Certified Reformation & Industrial Schools
		(8) Merchant Vessels & Lighthouses
		(9) Schools
	*/
	char * special[9] = {"Barracks", "H.M. Shipping", "Workhouses", "Hospitals",
		"Asylums", "Prisons", "Reform Schools", "Merchant Shipping", "Schools"};

	/*****************************/
	/* preset all trailing nulls */
	/*****************************/

	/* 4.7.0 change reg field to mean book number */
	book[4] = (char) NULL;
	household[6] = (char) NULL; strcpy(o_household, "XX");
	household_mem[4] = (char) NULL;
	parish[20] = (char) NULL; *o_parish = (char)NULL;
	/* ed[3] = (char) NULL; delayed */ strcpy(o_ed, "XX");
	folio[4] = (char) NULL;
	page[4] = (char) NULL; strcpy(o_page, "XX");
	/* sched[3] = (char) NULL; 4.4.8 delayed */
	house_num[4] = (char) NULL;
	house_name[30] = (char) NULL;
	surname[24] = (char) NULL;
	forenames[24] = (char) NULL;
	rel[6] = (char) NULL;
	age[3] = (char) NULL;
	occ[30] = (char) NULL;
	chap[3] = (char) NULL;
	place[20] = (char) NULL;
	dis[6] = (char) NULL;
	notes[44] = (char) NULL;
	parishe[20] = (char) NULL; *o_parishe= (char)NULL;
	altchap[3] = (char) NULL;
	altplace[20] = (char) NULL;

	nerr=0; nwarn = 0;

	/* set the read buffer and clear a bit before it
		Actually we set it to 0's as a cheat - it can improve the vote for the
		first record if it is chopped short */
	buf = (char *)realbuf + LEEWAY;

	memset(realbuf, '0', LEEWAY);

	/******************************/
	/* Decode the call parameters */
	/******************************/

	if (argc == 1)
	{
		/* If no arguments we call the usage routine and exit */
		usage();
		return 1;
	}

	/* handle the program options */
	i = options(argc,argv);

	/* If there is still no file name then what do we do? */
	if (i == 0)
		return 2;

	/* 4.5.4 get path of the binary for this program - the fname and ext are ignored */

	/* check size of path before using (dat2csv is only 7 characters, allow for 8) */
	if (strlen(argv[0]) > 126)
	{
		fprintf(stderr, "Fatal: Program path too long, move it closer to the root of the disk\n");
		return 12;
	}

	splitpath(argv[0], ppath, fname, ext);

	/* 4.5.3 check bounds of path before using */
	if (strlen(argv[i]) > 127)
	{
		fprintf(stderr, "Fatal: File path too long, move it closer to the root of the disk\n");
		return 12;
	}

	splitpath(argv[i], path, fname, ext);
	strcpy(input, argv[i]);

	/* Open the input and output files. Get the error file open asap. */

	strcpy(error, makepath(path, fname, "erd"));
	if ((err = fopen(error, "w")) == (FILE *)NULL)
	{
		fprintf(stderr, "Fatal: Messages file (%s) failed to open\n", error);
		perror("Fatal");
		return 12;
	}

	if (*ext == (char)NULL)
	{
		fprintf(err, "Fatal: Full file name required including the three letter extension - %s\n", input);
		fprintf(stderr, "Fatal: Full file name required including the three letter extension\n");
		fclose(err);
		return 12;
	}

	/* decode the file extension to work out whether file is checked or not */
	if (strcmp(ext, "dat") == 0 || strcmp(ext, "DAT") == 0)
		rlen = 256;
	else if (strcmp(ext, "vld") == 0 || strcmp(ext, "VLD") == 0)
		rlen = 299;
	else
		rlen = 276;

	/* decode the file name to work out what we are dealing with */

	ret = decode_census(fname, webname, proref);
	if (ret > 0)
	{
		fclose(err);
		return ret;
	}

	if ((mode & WEB) == WEB && rlen == 256)
	{
		/* Nowdays they should be validated (299) records but we also allow
			checked (276) to allow for the old VALCENS output files */
		fprintf(err,
			"Fatal: dat2csv does not support WEB format for unchecked records\n");
		fprintf(stderr,
			"Fatal: dat2csv does not support WEB format for unchecked records\n");
		fclose(err);
		return 12;
	}

	if ((mode & WEB) == WEB)
	{
		/* read in the Chapman code vs. County name list for WEB only */

		block = (char *)malloc(sizeof(char) * 24 * 150);
		ccode = (char * *)malloc(sizeof(char *) * 150);
		county = (char * *)malloc(sizeof(char *) * 150);

		/* a different file for 1841
		4.7.2 correct case */
		sprintf(temp, "CHAP%d", year > 1841 ? 1891 : 1841);
		/* 4.5.4 locate it in the binary directory
		4.7.2 correct case*/
		strcpy(temp, makepath(ppath, temp, "TXT"));

		if ((in = fopen(temp, "r")) == (FILE *)NULL)
		{
			fprintf(err, "Fatal: Aux file (%s) failed to open\n", temp);
			fprintf(stderr, "Fatal: Aux file (%s) failed to open\n", temp);
			perror("Fatal");
			fclose(err);
			return 12;
		}

		/* read lines of the form county[20]ccode[4], one per line */
		for (i = 0, cp = block; i < 150 && fgets(buf, 40, in) != (char *)NULL; i++, cp += 24)
		{
			county[i] = cp;
			strncpy(cp, buf, 23);
			ccode[i] = cp + 20;
			*(cp + 23) = (char)NULL;
			for (j = 19; buf[j] == ' '; --j)
				;
			*(cp + j + 1) = (char)NULL;
		}

		fclose(in);

		/* 4.4.9 check we don't overflow data area */
		if (i >= 150)
		{
			fprintf(err, "Fatal: Too many counties in aux file\n");
			fprintf(stderr, "Fatal: Too many counties in aux file\n");
			fclose(err);
			return 12;
		}

		ccode[i] = (char *)NULL;
	}

	/* Open the input and output files */

	if ((in = fopen(input, "rb")) == (FILE *)NULL)
	{
		fprintf(err, "Fatal: Input file (%s) failed to open\n", input);
		fprintf(stderr, "Fatal: Input file (%s) failed to open\n", input);
		perror("Fatal");
		fclose(err);
		return 12;
	}
	setvbuf(in, (char *)NULL, _IONBF, 0); /* no buffering */

	/* opening of the output file is delayed for WEB & BRIEF because we need
		information from the first record */
	if ((mode & (WEB + BRIEF)) == 0)
	{
		strcpy(output, makepath(path, fname, "csv"));
		if ((out = fopen(output, "w")) == (FILE *)NULL)
		{
			fprintf(err, "Fatal: Output file (%s) failed to open\n", output);
			fprintf(stderr, "Fatal: Output file (%s) failed to open\n", output);
			perror("Fatal");
			fclose(in);
			fclose(err);
			return 12;
		}
	}

	/****************************************/
	/* preset some values for counters etc. */
	/****************************************/

	if ((tod = time((time_t *)NULL)) != -1)
	{
		strcpy(date, asctime(gmtime(&tod)));
		/* get rid of the newline */
		i = strlen(date);
		date[i - 1] = (char)NULL;
		strcat(date, " UTC");
	}
	else
		*date = (char)NULL;

	fprintf(err, "Error listing for dat2csv (Rick Parsons) V%s %s converting %s\n"
"[When reporting program errors, be sure to send this file and the original data file]\n\n",
		VERSION, date, input);
	fprintf(err,
		"Census year = %d, input format = %s, output format = ",
		year, (rlen == 299) ? "VALD-REV" : ((rlen == 256) ? "IN-CENS" : "CHECK-CENS"));

	if ((mode & ALL) == ALL)
		fprintf(err, "ALL");
	if ((mode & BRIEF) == BRIEF)
		fprintf(err, "SS-CENS");
	if ((mode & WEB) == WEB)
		fprintf(err, "WEB");

	if (dstrip)
		fprintf(err, ", STRIP");

	fprintf(err, "%s\n\n", (quiet) ? ", QUIET" :((verbose) ? ", VERBOSE" : ""));

	if ((mode & BRIEF) == BRIEF)
		fprintf(err,
"Add 3 to record numbers in messages to get corresponding row in .CSV file.\n\n");

	if (dstrip && (mode & WEB) != WEB)
		fprintf(err,
"The correspondence between record numbers and rows will be distorted when records are deleted.\n\n");

	n = 0;
	foffset = 0;

	/*****************************************/
	/* big loop reading one record at a time */
	/*****************************************/

	while (TRUE)
	{
		i = fread(buf, sizeof(char), rlen, in);
		foffset += rlen;

		if (i != rlen && i != 0)
		{
			fprintf(err,
				"Error: Short record at end of file - possibly truncated\n"), nerr++;
			break;
		}

		if (feof(in))
			break;

		/* terminate the buffer */
		*(buf + rlen) = (char)NULL;

		/* here is where we start to get clever and try to re-allign mangled files.*/

		/*************************************************************/
		/* Check that the record is lined up (uses 8 bit characters) */
		/*************************************************************/
		if (vote((unsigned char *)buf) < MAXVOTE)
		{
			/* get another record's worth of data (but put the file position back) */
			fread(buf + rlen, sizeof(char), rlen, in);
			fseek(in, foffset, SEEK_SET);

			/* make sure that there is a still a null after the input buffer */
			*(buf + 2 * rlen) = (char)NULL;

			/* realbuf is LEEWAY characters before the actual buffer */
			v = maxvote(realbuf, rlen + LEEWAY);
			if (v < LEEWAY)
			{
				fprintf(err,
			"Error: record %d re-aligning by backspacing %d characters\n",
						n + 1, LEEWAY - v), nerr++;

				/* shift the record along (right) a bit to line it up */
				for (j = rlen - 1; j >= 0; --j)
					buf[j] = (char)realbuf[j + v];

				/* line up the file for next time */
				foffset -= (LEEWAY - v);
				fseek(in, foffset, SEEK_SET);
			}
			else if (v > LEEWAY)
			{
				fprintf(err,
					"Error: record %d re-aligning by skipping %d characters\n",
					n + 1, v - LEEWAY), nerr++;

				/* shift the record along (left) a bit to line it up */
				for (j = 0; j < rlen; j++)
					buf[j] = (char)realbuf[j + v];

				/* line up the file for next time */
				foffset -= (LEEWAY - v);
				fseek(in, foffset, SEEK_SET);
			}
			/* if no better allignment, then continue with what we have */
		}

		/* terminate the buffer again */
		*(buf + rlen) = (char)NULL;

		/* anticipating potential trouble next time, put part of the end of this
			record in front so that it is available for the vote next time around */
		memcpy(realbuf, buf + rlen - LEEWAY, LEEWAY);

		/* provide a rolling count so we can see that something is happening */
		n++;
		fprintf(stderr, "%d\r", n);

		/*****************************/
		/* Check that it is complete.*/
		/*****************************/

		/* first see if there is a spurious CRLF in the record */
		if ((cp = strchr(buf + b_parish, '\r')) != (char *)NULL)
		{
			fprintf(err, "Error: Carriage-Return detected in record %d\n",
				n), nerr++;
			/* Blank out record from that point on */
			/* memset(cp, ' ', (int)(buf - cp + rlen)); */
			/* try less drastic means - just elide the offending characters */
			*cp = ' ';

			/* Line up the file ready for the next record at the newline.
				Strictly this is not correct, but lusers who edit the file
				tend to miss out the first character of the next record as
				well. */
			foffset = foffset - rlen + 1 + (cp - buf);
			fseek(in, foffset, SEEK_SET);
		}

		/* 4.7.0 change reg field to mean book number */
		/* early IN-CENS files can have null book field (it was supposed
			to be Registration District then).
			This is complex to allow for all cases and bad data. */
		if (year == 1891)
			memcpy(book, buf + b_book, 4);
		else
			strncpy(book, buf + b_book, 4);
		strip(book);
		/* 4.7.0 blank out zero book numbers
		if (atoi(book) == 0)
			*book = (char)NULL;
*/

		/* for safety, make sure that there are no non-printables in the record */
		t = verbose;
		for (j = 0; j < rlen; j++)
			/* 4.4.11 cast to unsigned so that accented characters get through */
			if ((unsigned char)buf[j] < ' ')
			{
				if (t && j > b_household)
					fprintf(err,
						"Info: record %d non printable characters\n", n), t = FALSE;
				buf[j] = ' ';
			}

		/************************************/
		/* extract the data from the record */
		/************************************/

		ed[3] = (char) NULL; /* limit initial ED to 3 characters */
		sched[3] = (char) NULL; /* 4.4.8 limit initial sched to 3 characters */

		/* 4.7.0 book number was extracted above */
		strncpy(household, buf + b_household, 6); strip(household);
		strncpy(household_mem, buf + b_household_mem, 4); strip(household_mem);
		strncpy(parish, buf + b_parish, 20); strip(parish);
		strncpy(ed, buf + b_ed, 3); strip(ed);
		strncpy(folio, buf + b_folio, 4); strip(folio);
		strncpy(page, buf + b_page, 4); strip(page);
		strncpy(sched, buf + b_sched, 3); strip(sched);
		strncpy(house_num, buf + b_house_num, 4); strip(house_num);
		strncpy(house_name, buf + b_house_name, 30); strip(house_name);
		strncpy(surname, buf + b_surname, 24); strip(surname);
		strncpy(forenames, buf + b_forenames, 24); strip(forenames);
		strncpy(rel, buf + b_rel, 6); strip(rel);
		strncpy(age, buf + b_age, 3); strip(age);
		strncpy(occ, buf + b_occ, 30); strip(occ);
		strncpy(chap, buf + b_chap, 3); strip(chap);
		strncpy(place, buf + b_place, 20); strip(place);
		strncpy(dis, buf + b_dis, 6); strip(dis);
		strncpy(notes, buf + b_notes, 44); strip(notes);

		if (rlen > 256) /* extra field in Check_CENS files */
		{
			strncpy(parishe, buf + b_parishe, 20); strip(parishe);
		}

		if (rlen > 276) /* extra fields in Vald-rev files */
		{
			strncpy(altchap, buf + b_altchap, 3); strip(altchap);
			strncpy(altplace, buf + b_altplace, 20); strip(altplace);
		}

		/*******************************************/
		/* There are now some data validity checks */
		/*******************************************/

		/* 4.6.6 For the first record ony, set the Welsh flag iff the language
			field is non blank 4.6.7 but not scottish */
		if (n == 1 && ! sct && year >= 1881 && buf[b_welsh] != ' ')
			wls = TRUE;

		/* each has its own quiet/verbose check to make it easier to change */

		if (buf[b_del] == 'D')
		{
			if (! quiet)
				fprintf(err, "Warning: record %d: Marked deleted", n), nwarn++;

			/* If we are stripping out deleted records, then we are done */
			if (dstrip)
			{
				if (! quiet)
					fprintf(err, " - removed\n");
				continue;
			}

			if ((mode & BRIEF) == BRIEF)
				strcpy(notes, "DELETED");
			if (! quiet)
				fprintf(err, "\n");
		}
		/* 4.6.8 Change a blank deletion marker to 'Z'. This is to stop spreadsheet
			programs removing totally empty columns from the left hand side of a
			file in CSV FULL mode. */
		else if (buf[b_del] == ' ')
			buf[b_del] = 'Z';

		if (! quiet && ! numeric(household))
			fprintf(err,
				"Warning: record %d, Household number = <%s> not numeric\n",
				n, household), nwarn++;
		if (! quiet && ! numeric(household_mem))
			fprintf(err,
				"Warning: record %d, Household member number = <%s> not numeric\n",
				n, household_mem), nwarn++;

		/* 4.7.0 see if book numbers are in use */
		/* 4.7.4 Allow non numeric ones */
		if (n == 1 && year == 1841 && !sct)
		{
			if (atoi(book) > 100)
			{
				if (!quiet)
					fprintf(err, "Warning: Old format file, Book numbers not in use\n"),
						nwarn++;
			}
			else if (*book == (char)NULL || strcmp(book, "0") == 0)
			{
				if (!quiet)
					fprintf(err, "Warning: Book numbers not set\n"),
						nwarn++;
			}
			else
				gbook = TRUE;
		}

		/* 4.7.0 Make sure we have a valid potential book number field */
		/* 4.7.4 Allow non-numeric ones */
		/* if (year == 1841 && !sct && !numeric(book))
			*book = (char)NULL;
		*/

		/* check that we are starting at the beginning */
		if (n == 1 && rlen == 256)
		{
			if (strcmp(household, "000001"))
				fprintf(err, "Error: record 1, household number = %s\n",
					household), nerr++;
			if (! quiet && strcmp(household_mem, "0001"))
				fprintf(err, "Warning: record 1, household member number = %s\n",
					household_mem), nwarn++;
		}

		/* A curiosity of checked/validated pieces is that the household numbers
			sometimes, but not always have leading zeros. Remove them */
		if (rlen > 256)
		{
			sprintf(household, "%d", atoi(household));
			sprintf(household_mem, "%d", atoi(household_mem));
		}

		/* this is to trap a suspected bug in *-CENS */
		if (atoi(household) == 0)
			fprintf(err, "Error: record %d, household number = 0\n", n), nerr++;
		if (verbose && atoi(household_mem) == 0)
			fprintf(err, "Info: record %d, household member number = 0\n",
				n);

		if (verbose && *parish == (char)NULL)
			fprintf(err,
				"Info: record %d, Civil parish missing\n", n);

		if (! quiet && ! numeric(ed))
			fprintf(err,
				"Warning: record %d, ED = <%s> not numeric\n", n, ed);

		if (buf[b_ed_a] == '-' || buf[b_ed_a] == 'x' || buf[b_ed_a] == 'X')
		{
			if (verbose)
				fprintf(err, "Info: record %d, ED suffix = <%c>\n",
					n, buf[b_ed_a]);
			buf[b_ed_a] = ' ';
		}
		else if (! quiet && buf[b_ed_a] != ' ' && ! isalnum(buf[b_ed_a]))
			fprintf(err, "Warning: record %d, ED suffix = <%c> not alphanumeric\n",
				n, buf[b_ed_a]), nwarn++;

		if (verbose && ! numeric(folio))
			fprintf(err,
				"Info: record %d, Folio number = <%s> not numeric\n",
				n, folio);

		if (buf[b_folio_a] == '-' || buf[b_folio_a] == 'x' || buf[b_folio_a] == 'X')
		{
			if (verbose)
				fprintf(err, "Info: record %d, folio suffix = <%c>\n",
					n, buf[b_folio_a]);
			buf[b_folio_a] = ' ';
		}
		else if (! quiet && buf[b_folio_a] != ' ' && ! isalpha(buf[b_folio_a]))
			fprintf(err, "Warning: record %d, Folio suffix = <%c> not alphabetic\n",
				n, buf[b_folio_a]), nwarn++;

		if (verbose && ! numeric(page))
			fprintf(err,
				"Info: record %d, Page number = <%s> not numeric\n", n, page);

		if (*sched == (char)NULL)
		{
			if (year > 1841 && ! quiet)
				fprintf(err, "Warning: record %d, Schedule number is blank\n", n),
					nwarn++;
			strcpy(sched, "0");
		}
		else if (strcmp(sched, "-") == 0)
		{
			if (verbose)
				fprintf(err, "Info: record %d, Schedule number = -\n", n);
			*sched = '0';
		}
		else if (verbose && ! numeric(sched))
			fprintf(err,
				"Info: record %d, Schedule number = <%s> not numeric\n", n, sched);

		/* 4.4.8 here is the hack to allow for schedules > 999 */
		if (buf[b_sched_a] == '!')
		{
			sprintf(sched, "%d", atoi(sched) + 1000);
			buf[b_sched_a] = ' ';
		}

		/* a zero schedule cannot have an alpha suffix */
		if (strcmp(sched, "0") == 0)
			buf[b_sched_a] = ' ';

		if (buf[b_sched_a] == '-' || buf[b_sched_a] == 'x' || buf[b_sched_a] == 'X')
		{
			if (verbose)
				fprintf(err, "Info: record %d, Schedule number suffix = <%c>\n",
					n, buf[b_sched_a]);
			buf[b_sched_a] = ' ';
		}
		else if (! quiet && buf[b_sched_a] != ' ' && ! isalpha(buf[b_sched_a]))
			fprintf(err,
				"Warning: record %d, Schedule number suffix = <%c> not alphabetic\n",
				n, buf[b_sched_a]), nwarn++;

		if (strcmp(house_num, "0") == 0 || strcmp(house_num, "-") == 0)
		{
			if (verbose)
				fprintf(err, "Info: record %d, house number = %s\n",
					n, house_num);
			*house_num = (char)NULL;
		}
		else if (verbose && ! numeric(house_num))
			fprintf(err,
				"Info: record %d, House number = <%s> not numeric\n",
				n, house_num);

		if (buf[b_house_num_a] == '-' || buf[b_house_num_a] == 'x'
			|| buf[b_house_num_a] == 'X')
		{
			if (verbose)
				fprintf(err, "Info: record %d, House number suffix = <%c>\n",
					n, buf[b_house_num_a]);
			buf[b_house_num_a] = ' ';
		}
		else if (verbose && buf[b_house_num_a] != ' ' && ! isalpha(buf[b_house_num_a]))
			fprintf(err, "Info: record %d, House number suffix = <%c> not alphabetic\n",
				n, buf[b_house_num_a]);

		if (buf[b_buvn] == ' ')
		{
			if (verbose)
				fprintf(err,
					"Info: record %d, Building/Unoccupied flag = < >\n",
					n);
			buf[b_buvn] = '-';
		}

		/* Remove any other garbage from unoccupied houses (CODE REMOVED) */
		if (buf[b_buvn] == 'b' || buf[b_buvn] == 'u' || buf[b_buvn] == 'v'
			|| buf[b_buvn] == 'n')
		{
		/*	*forenames = (char)NULL;
			*rel = (char)NULL;
			buf[b_cond] = ' ';
			buf[b_sex] = '-';
			*age = (char)NULL;
			buf[b_age_unit] = ' ';
			*occ = (char)NULL;
			*chap = (char)NULL;
			*place = (char)NULL;
			*dis = (char)NULL; */
		}
		else
		{

		if (verbose && buf[b_buvn] != '-' && buf[b_buvn] != 'x')
			fprintf(err,
				"Info: record %d, building/unoccupied flag = <%c> invalid\n",
				n, buf[b_buvn]);

		if (verbose && *surname == (char)NULL)
			fprintf(err,
				"Info: record %d, surname missing\n", n);

		if (verbose && *forenames == (char)NULL)
			fprintf(err,
				"Info: record %d, forename(s) missing\n", n);

		if (verbose && buf[b_name_err] != '-' && buf[b_name_err] != 'x')
			fprintf(err,
				"Info: record %d, name flag = <%c> invalid\n",
				n, buf[b_name_err]);

		if (verbose && year > 1841 && *rel == (char)NULL)
			fprintf(err,
				"Info: record %d, relationship missing\n", n);

		if (verbose && year > 1841 && buf[b_cond] != 'M' && buf[b_cond] != 'S'
			&& buf[b_cond] != 'U' && buf[b_cond] != 'W'
			&& buf[b_cond] != '-')
			fprintf(err,
				"Info: record %d, condition = <%c> invalid\n", n, buf[b_cond]);

		if (verbose && buf[b_sex] != 'M' && buf[b_sex] != 'F'
			&& buf[b_sex] != '-')
			fprintf(err,
				"Info: record %d, sex = <%c> invalid\n", n, buf[b_sex]);

		if (verbose && *age == (char)NULL)
			fprintf(err,
				"Info: record %d, age missing\n", n);
		else if (verbose && ! numeric(age))
			fprintf(err,
				"Info: record %d, age = <%s> not numeric\n", n, age);

		if (verbose && buf[b_age_unit] != 'y' && buf[b_age_unit] != 'm'
			&& buf[b_age_unit] != 'w' && buf[b_age_unit] != 'd'
			&& buf[b_age_unit] != '-')
			fprintf(err,
				"Info: record %d, age unit = <%c> invalid\n", n, buf[b_age_unit]);

		if (verbose && buf[b_detail_err] != '-' && buf[b_detail_err] != 'x')
			fprintf(err,
				"Info: record %d, detail flag = <%c> invalid\n",
				n, buf[b_detail_err]);

		if (verbose && buf[b_occ_err] != '-' && buf[b_occ_err] != 'x')
			fprintf(err,
				"Info: record %d, occupation flag = <%c> invalid\n",
				n, buf[b_occ_err]);

		if (verbose && *chap == (char)NULL)
			fprintf(err,
				"Info: record %d, county code missing\n", n);

		if (verbose && year > 1841 && *place == (char)NULL)
			fprintf(err,
				"Info: record %d, birth place missing\n", n);

		if ((mode & WEB) == WEB && rlen == 299)
		{
			/* if the birth place has been changed during validation then
				the original ones are of no use to us for WEB output.
				Copy them across. In practice either both or neither are set. */
			if (*altchap != (char)NULL)
				strcpy(chap, altchap);
			if (*altplace != (char)NULL)
				strcpy(place, altplace);
		}

		if (verbose && buf[b_place_err] != '-' && buf[b_place_err] != 'x')
			fprintf(err,
				"Info: record %d, birth place flag = <%c> invalid\n",
				n, buf[b_place_err]);

		/* 4.6.6 correctly check the Welsh language field */
		if (wls) /* only set if year >= 1881 */
		{
			if (verbose && buf[b_welsh] != 'W' && buf[b_welsh] != 'E'
				&& buf[b_welsh] != 'B' && buf[b_welsh] != '-')
				fprintf(err,
					"Info: record %d, Language marker = <%c> invalid\n",
					n, buf[b_welsh]);
		}
		/* 4.6.7 do it for Gaelic as well */
		else if (sct && year >= 1881)
		{
			if (verbose && buf[b_welsh] != 'G' && buf[b_welsh] != 'B'
				&& buf[b_welsh] != '-')
				fprintf(err,
					"Info: record %d, Language marker = <%c> invalid\n",
					n, buf[b_welsh]);
		}
		else if (verbose && year >= 1881 && buf[b_welsh] != ' ')
				fprintf(err,
					"Warning: record %d, Unexpected Language marker = <%c>\n",
					n, buf[b_welsh]);
		}

		/* 4.6.14 Re-instate Eccl districts in 1851 */
		if (verbose && rlen > 256 && year > 1841 && *parishe == (char)NULL)
			fprintf(err,
				"Info: record %d, Eccl District missing\n", n);

		/* numeric ED extension is a special case -> institution or shipping */
		if ((mode & WEB) != WEB && isdigit(buf[b_ed_a]))
			strcat(ed, "#");

		/**********************************/
		/* now print them out (WEB style) */
		/**********************************/

		if ((mode & WEB) == WEB)
		{	/* Open output file and print header on recieving first record */
		/* (back off code indentation here to keep it readable) */
		if (n == 1)
		{
			if (con)
				/* 4.7.0 change reg field to mean book number - this code is now obsolete */
				strcat(webname, book);

			strcpy(output, makepath(path, webname, "html"));
			if ((out = fopen(output, "w")) == (FILE *)NULL)
			{
				fprintf(err, "Fatal: Output file (%s) failed to open\n",
					output);
				fprintf(stderr, "Fatal: Output file (%s) failed to open\n",
					output);
				perror("Fatal");
				fclose(in);
				fclose(err);
				return 12;
			}

			/* 4.7.0 change reg field to mean book number */
			ret = webhead(out, input, proref, book, parish);
			if (ret > 0)
			{
				fclose(out); fclose(in); fclose(err);
				return ret;
			}
		}

		/* New Enumeration District started
		4.6.13 or (civil) parish - relevent to 1841 where consecutive EDs may
			be the same */
		if (strcmp(ed, o_ed) != 0 || buf[b_ed_a] != o_ed_a ||
			strcmp(parish, o_parish) != 0)
		{
			if (!firstp) /* we need to terminate the previous data */
				fprintf(out, "</pre>\n");

			firstp = TRUE;

			/* 4.6.5 add the id except on continuation EDs.
			   4.6.9 also omit from 1841 due to possible duplication.
				If there is more than one special per ED then duplicate ids will
				be produced. This is a bug but will occur rarely. */
			if (year != 1841 && (buf[b_ed_a] != ' ' || strcmp(ed, o_ed) != 0))
			/* 4.6.11 Supress the trailing space */
				fprintf(out, "<h3 id=\"ED%s%s\">", ed,
					squash(buf[b_ed_a], (char)NULL));
			/* Use book number for id in 1841 - this doesn't work
			else if (gbook)
				fprintf(out, "<h3 id=\"BK%s\">", book); */
			else
				fprintf(out, "<h3>");

			/* don't print zero EDs */
			if (atoi(ed) != 0)
				fprintf(out, "Enumeration District %s", ed);

			/* numeric ED extension is a special case -> institution or shipping */
			if (isdigit(buf[b_ed_a]))
				fprintf(out, " %s</h3>\n", special[buf[b_ed_a] - '1']);
			/* 4.6.13 In 1841 EDs don't continue over parish boundaries */
			else if (buf[b_ed_a] == ' ' && strcmp(ed, o_ed) == 0 &&
				(year != 1841 || strcmp(parish, o_parish) == 0))
				/* this caters for all but things like 3a (cont) which I can't detect */
				/* 4.7.1 remove spurious </a> */
				fprintf(out, " (Continued)</h3>\n");
			else
				fprintf(out, "%c</h3>\n", buf[b_ed_a]);

			/* 4.6.2 */
			if (ipart == 0)
			{
				fprintf(out, "<p class=\"big\">Civil Parish of %s", parish);
				/* 4.6.14 re-instate Eccl districts in 1851 */
				if (year > 1841)
					fprintf(out, ", Eccl. District of %s", parishe);
				fprintf(out, "</p>\n");
			}

			strcpy(o_ed, ed);
			o_ed_a = buf[b_ed_a];
			/* 4.6.13 */
			strcpy(o_parish, parish);

			/* force a reprint of the page/folio numbers - this is just in case we
				have a change of ED in the middle of a page. This can only happen
				if there are institutions included in the standard ED books */
			o_folio_a = 'x';
		}

		if (strcmp(page, o_page) != 0
			|| strcmp(folio, o_folio) != 0 || buf[b_folio_a] != o_folio_a)
		{	/* new page started - the folio check is to allow
				for shipping where page=0 */
			if (!firstp) /* we need to terminate the previous data */
				fprintf(out, "</pre>\n");
			else
			{
				firstp = FALSE;
			}

			fprintf(out, "<h4>");
			/* 4.7.0 Print out book numbers if we have them */
			if (gbook)
				fprintf(out, "Book %s ", book);
			/* SCT doesn't require folio either but we don't check that yet */
			if (*folio != (char)NULL)
				fprintf(out, "Folio %s%c", folio, buf[b_folio_a]);
			if (strcmp(page, "0") != 0)
				fprintf(out, " Page %s", page);
			fprintf(out, "</h4>\n<pre class=\"data\">\n");
			strcpy(o_page, page);
			strcpy(o_folio, folio);
			o_folio_a = buf[b_folio_a];
		}

		if (strcmp(household, o_household) != 0)
		{	/* new household started */
			if (year > 1841)
			{
				if (strcmp(sched, "0") != 0)
					/* 4.6.10 Suppress spurious commas */
					fprintf(out, "%s", sprep(sched, buf[b_sched_a]));
				fprintf(out, ",");
			}

			/* 4.6.12 1841 can have house numbers (why not?)
				we have to get smart here because the House
				number/name is being output as one field,
				so if there is a comma, the WHOLE must be
				enclosed in quotes */

			strcpy(work, house_num);
			strcat(work, squash(buf[b_house_num_a], (char)NULL));
			if (strcmp(house_num, "") != 0)
				strcat(work, " ");

			strcat(work, house_name);

			/* 1851 didn't have an uninhabited column so put the information
				into the house name field */
			if (year == 1851 && (buf[b_buvn] == 'b' || buf[b_buvn] == 'u'
				|| buf[b_buvn] == 'v' || buf[b_buvn] == 'n'))
			{
				if (*work != (char)NULL)
					strcat(work, ", ");
				strcat(work, surname);
			}

			fprintf(out, "%s,", prepare(work, buf[b_buvn]));

			/* it there is an error flag, then lose it for web printing */
			if (buf[b_buvn] == 'x')
				buf[b_buvn] = ' ';

			if (year != 1851)
				fprintf(out, "%s,", squash(buf[b_buvn], '1'));

			/* strcpy(o_sched, sched); */
			strcpy(o_household, household);
		}
		else if (year > 1851)
			fprintf(out, ",,,");
		else
			fprintf(out, ",,");

		if (buf[b_buvn] == 'b' || buf[b_buvn] == 'u' || buf[b_buvn] == 'v'
			|| buf[b_buvn] == 'n')
			/* strcpy(work, surname);
			Remove descriptive field for the WEB format */
			*work = (char)NULL;
		else
		{
			/* Revert the surname to mixed case 4.5.1 change to 8 bit characters */
			for (ucp = (unsigned char *)surname + 1; *ucp != (unsigned char)NULL; ucp++)
				*ucp = tolower8(*ucp);

			strcpy(work, forenames);
			strcat(work, " ");
			strcat(work, surname);
		}

		fprintf(out, "%s,", prepare(work, buf[b_name_err]));
		if (year > 1841)
		{
			fprintf(out, "%s,", prepare(rel, buf[b_detail_err]));
			fprintf(out, "%s,", prepare(squash(buf[b_cond], (char)NULL), buf[b_detail_err]));
		}

		/* blank out the 999y ages */
		if (strcmp(age, "999") == 0)
		{
			*age = (char)NULL;
			buf[b_age_unit] = ' ';
		}

		strcpy(work, age);
		if (buf[b_age_unit] != 'y')
			strcat(work, squash(buf[b_age_unit], 0));

		/* split the ages back into two columns, male and female */

		if (buf[b_sex] == 'F')
			fprintf(out, ",");
		fprintf(out, "%s", prepare(work, buf[b_detail_err]));
		fprintf(out, ",");
		if (buf[b_sex] != 'F')
			fprintf(out, ",");

		/* Try to detect the IN-CENS style employee etc. flags and extract them */
		if (year > 1881)
		{
			c = decode_emp(occ);
			switch (c)
			{
				case 'E': strcpy(emp, "Employed");
					break;
				case 'R': strcpy(emp, "Employer");
					break;
				case 'N': strcpy(emp, "Neither");
					break;
				default: *emp = (char)NULL;
			}
		}

		fprintf(out, "%s,", prepare(occ, buf[b_occ_err]));
		if (year > 1881)
		{
			fprintf(out, "%s,", prepare(emp, buf[b_occ_err]));
		}

		/* convert the Chapman code back to a county name */
		p = cfind(ccode, chap);

		/* same sort of game as house name for the place of birth */
		*work = (char)NULL;
		if (year > 1841)
		{
			strcpy(work, place);
			strcat(work, " ");
		}

		strcat(work, p >= 0? county[p]: chap);

		fprintf(out, "%s", prepare(work, buf[b_place_err]));

		if (year > 1841)
			fprintf(out, ",%s", prepare(dis, 0));

		/* 4.6.7 Print Language flag */
		if (year >= 1881 && (wls || sct))
		{
			fprintf(out, ",%c", buf[b_welsh]);
		}

		fprintf(out, ",%s\n", prepare(notes, 0));

		}

		/*******************/
		/* BRIEF style CSV */
		/*******************/

		else if ((mode & BRIEF) == BRIEF)
		{	/* Open output file and print header on recieving first record */

		if (n == 1)
		{
			strcpy(output, makepath(path, fname, "csv"));
			if ((out = fopen(output, "w")) == (FILE *)NULL)
			{
				fprintf(err,
					"Fatal: Output file (%s) failed to open\n", output);
				fprintf(stderr,
					"Fatal: Output file (%s) failed to open\n", output);
				perror("Fatal");
				fclose(in);
				fclose(err);
				return 12;
			}

			/* 4.7.0 change reg field to mean book number */
			fprintf(out, "%s,%s,Output by dat2csv (Rick Parsons) V%s %s from %s\n",
				proref, book, VERSION, date, input);
			fprintf(out, "Civil Parish");
			if (rlen > 256)
				fprintf(out, ",Eccl. District");
			fprintf(out, ",ED,Folio,Page,Schd,House,Address,X,"
				"Surname,Forenames,X,Rel.,C,Sex,Age,X,"
				"Occupation,E,X,CHP,Place of birth,X");
			if (rlen > 276)
				fprintf(out, ",ALT,Alt. Place");
			/* 4.7.0. Add a book column to the end of validated 1841 files */
			fprintf(out, ",Dis.,L,Notes");
			if (rlen > 276 && year == 1841 && !sct)
				fprintf(out, ",Book");
			fprintf(out, "\n");

			fprintf(out, "abcdefghijklmnopqrst");
			if (rlen > 256)
				fprintf(out, ",abcdefghijklmnopqrst");
			fprintf(out, ",###a,####a,####,###a,####a,"
				"abcdefghijklmnopqrstuvwxyzabcd,X,abcdefghijklmnopqrstuvwx,"
				"abcdefghijklmnopqrstuvwx,X,abcdef,C,S,###a,X,"
				"abcdefghijklmnopqrstuvwxyzabcd,E,X,abc,abcdefghijklmnopqrst,X");
			if (rlen > 276)
				fprintf(out, ",abc,abcdefghijklmnopqrst");
			/* 4.7.0. Add a book column to the end of validated 1841 files */
			fprintf(out,
				",abcdef,L,abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqr");
			if (rlen > 276 && year == 1841 && !sct)
				fprintf(out, ",####");
			fprintf(out, "\n");
		}

		/* New page started? Columns A to D (extra one for checked pieces) */
		if (strcmp(page, o_page) != 0
			|| strcmp(folio, o_folio) != 0 || buf[b_folio_a] != o_folio_a
			|| strcmp(ed, o_ed) != 0 || buf[b_ed_a] != o_ed_a
			|| strcmp(parish, o_parish) != 0
			|| (rlen > 256 && strcmp(parishe, o_parishe) != 0))
		{	/* new page started - the folio check is to allow for shipping
				where page=0 the ED check is for intitutions that start
				in the middle of a page */
			fprintf(out, "%s", prepare(parish, 0));
			if (rlen > 256)
				fprintf(out, ",%s", prepare(parishe, 0));
			/* 4.6.10 allow for spurious commas in ed, folio, page */
			fprintf(out, ",%s", sprep(ed, buf[b_ed_a]));
			fprintf(out, ",%s", sprep(folio, buf[b_folio_a]));
			fprintf(out, ",%s", prepare(page, 0));
/* 4.7.0 Moved later so the test can be used again
			strcpy(o_page, page);
*/
			strcpy(o_folio, folio);
			o_folio_a = buf[b_folio_a];
			strcpy(o_ed, ed);
			o_ed_a = buf[b_ed_a];
			strcpy(o_parish, parish);
			if (rlen > 256)
				strcpy(o_parishe, parishe);
		}
		else if (rlen > 256)
			fprintf(out, ",,,,");
		else
			fprintf(out, ",,,");

		if (strcmp(household, o_household) != 0)
		{	/* new household started - Columns E to H */
			/* Note: address changes during a household should not
				occur and are not reflected in the output */
			*work = (char)NULL;

			/* 1841 scedules do not exist, but force them so that we can see the
				household breaks */
			if (year == 1841)
				strcpy(sched, "0");
			/* 4.6.10 allow for spurious commas in sched, house number */
			fprintf(out, ",%s", sprep(sched, buf[b_sched_a]));
			fprintf(out, ",%s", sprep(house_num, buf[b_house_num_a]));

			fprintf(out, ",%s,%c", prepare(house_name, 0), buf[b_buvn]);

			strcpy(o_household, household);
		}
		else
			fprintf(out, ",,,,");

		fprintf(out, ",%s", prepare(surname, 0));
		fprintf(out, ",%s,%c", prepare(forenames, 0), buf[b_name_err]);
		fprintf(out, ",%s,%c", prepare(rel, 0), buf[b_cond]);
		/* 4.6.10 allow for spurious commas in age */
		fprintf(out, ",%c", buf[b_sex]);
		fprintf(out, ",%s", sprep(age, buf[b_age_unit]));
		fprintf(out, ",%c", buf[b_detail_err]);

		/* Try to detect the IN-CENS style employee etc. flags and extract them */
		if (year > 1881)
		{
			c = decode_emp(occ);
			emp[0] = c;
			emp[1] = (char)NULL;
		}
		else
			emp[0] = (char)NULL;

		fprintf(out, ",%s,%s,%c", prepare(occ, 0), emp, buf[b_occ_err]);
		/* 4.6.10 allow for spurious commas in the Chapman codes */
		fprintf(out, ",%s", prepare(chap, 0));
		fprintf(out, ",%s,%c", prepare(place, 0), buf[b_place_err]);

		if (rlen > 276)
		{
			fprintf(out, ",%s", prepare(altchap, 0));
			fprintf(out, ",%s", prepare(altplace, 0));
		}

		fprintf(out, ",%s,%c", prepare(dis, 0), buf[b_welsh]);
		/* 4.7.0. Add a book column to the end of validated 1841 files */
		fprintf(out, ",%s", prepare(notes, 0));
		if (rlen > 276 && year == 1841 && !sct && strcmp(page, o_page) != 0)
			fprintf(out, ",%s", prepare(book, 0));
		fprintf(out, "\n");
/* 4.7.0 Moved from above */
		strcpy(o_page, page);
		}

		/*****************/
		/* ALL style CSV */
		/*****************/

		else /* ALL */
		{

		/* 4.7.0 change reg field to mean book number - prepare it for printing */
		fprintf(out, "%c,%s,%s,%s,",
			buf[b_del], prepare(book, 0), household, household_mem);
		fprintf(out, "%s,", prepare(parish, 0));
		/* 4.6.10 allow for spurious commas in ed, folio, page, sched, house num */
		fprintf(out, "%s,", sprep(ed, buf[b_ed_a]));
		fprintf(out, "%s,", sprep(folio, buf[b_folio_a]));
		fprintf(out, "%s,", prepare(page, 0));
		fprintf(out, "%s,", sprep(sched, buf[b_sched_a]));
		fprintf(out, "%s,", sprep(house_num, buf[b_house_num_a]));
		fprintf(out, "%s,%c,", prepare(house_name, 0), buf[b_buvn]);
		fprintf(out, "%s,", prepare(surname, 0));
		fprintf(out, "%s,%c,", prepare(forenames, 0), buf[b_name_err]);
		fprintf(out, "%s,%c,", prepare(rel, 0), buf[b_cond]);
		/* 4.6.10 allow for spurious commas in age */
		fprintf(out, "%c,", buf[b_sex]);
		fprintf(out, "%s,", sprep(age, buf[b_age_unit]));
		fprintf(out, "%c,", buf[b_detail_err]);
		fprintf(out, "%s,%c,", prepare(occ, 0), buf[b_occ_err]);
		/* 4.6.10 allow for spurious commas in the Chapman codes */
		fprintf(out, "%s,", prepare(chap, 0));
		fprintf(out, "%s,%c,", prepare(place, 0), buf[b_place_err]);
		fprintf(out, "%s,%c,", prepare(dis, 0), buf[b_welsh]);
		fprintf(out, "%s", prepare(notes, 0));
		if (rlen > 256)
			fprintf(out, ",%s", prepare(parishe, 0));
		if (rlen > 276)
		{
			fprintf(out, ",%s", prepare(altchap, 0));
			fprintf(out, ",%s", prepare(altplace, 0));
		}
		fprintf(out, "\n");

		}
	}

	if ((mode & WEB) == WEB)
	{
		fprintf(out, "</pre>");
		/* 4.7.0 change reg field to mean book number */
		webfoot(out, input, proref, book, parish);
		free(county);
		free(ccode);
		free(block);
	}

	fprintf(err, "\nTranslated %d records\n", n);
	/* 4.6.1 */
	fprintf(err, "\nOutput file is %s\n", output);

	fprintf(stderr, "\nTranslated %d records\n", n);
	fprintf(stderr, "\n %d errors", nerr);
	if (! quiet)
		fprintf(stderr, ", %d warnings", nwarn);
	fprintf(stderr, ". see %s\n", error);

	fclose(err);

	/* don't close file if it hasn't been opened */
	if (n > 0 || (mode & (WEB + BRIEF)) == 0)
		fclose(out);
	fclose(in);

	if (nerr > 0)
		return 8;
	if (nwarn > 0)
		return 4;

	return 0;
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash (unix) or a slash (DOS) */
int options(int argc,char *argv[])
{
	int i,firstnonoption=0;
	extern int mode, verbose, quiet, dstrip;

	mode = BRIEF;
	quiet = FALSE;
	verbose = FALSE;
	dstrip = FALSE;

	for (i=1; i < argc;i++)
	{
		if (
#ifndef UNIX
			argv[i][0] == '/' ||
#endif
			argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
				case 'a':
				case 'A':
					mode = ALL;
					break;
				case 'b':	/* for backwards compatability */
				case 'B':
					mode = BRIEF;
					break;
				case 'h':
				case 'H':
				case '?':
					usage();
					return 0;
				case 'q':
				case 'Q':
					quiet = TRUE;
					verbose = FALSE;
					break;
				case 'v':
				case 'V':
					quiet = FALSE;
					verbose = TRUE;
					break;
				case 'w':
				case 'W':
					mode = WEB;
					quiet = FALSE;	/* Forced verbosity */
					dstrip = TRUE;	/* don't put out deleted lines */
					verbose = TRUE;
					break;
				case 'z':
				case 'Z':
					dstrip = TRUE;
					break;
				default:
					fprintf(stderr,"unknown option %s\n",argv[i]);
					break;
			}
		}
		else
		{
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

void strip(char * string)
{	/* Subroutine to remove leading and trailing blanks from a field */
	int l, i;

	/* strip off leading blanks */
	for (i = 0; string[i] == ' '; i++)
		;
	if (i != 0)
	{
		l = 0;
		while (string[i] != (char)NULL)
			string[l++] = string[i++];
		string[l] = (char)NULL;
	}

	/* now strip off trailing blanks */
	l = strlen(string);
	for(i = --l; i >= 0; --i)
		if (string[i] != ' ')
			break;
	string[i + 1] = (char) NULL;
	/* remove trailing commas */
	if (i >= 0 && string[i] == ',')
		string[i] = (char)NULL;

	return;
}

/* this routine uses a common work area so must not be
	called twice in the same expression */
char * prepare(char * string, char flag)
{	/* Subroutine to enclose a string in quotes if it is needed, CSV must have fields
		enclosed in quotes if they contain a comma or other quotes
		it also does a few extra things for WEB output to avoid HTML glitches
		and highlight flagged fields */
	extern char hold[];
	char work[50];
	extern int mode;
	char * ip, * jp;

	hold[0] = (char)NULL;
	/* for web output, convert null markers (hyphen) to a blank field */
	if((mode & WEB) == WEB && strcmp(string, "-") == 0)
		return hold;

	jp = hold;

	if (strchr(string, ',') || strchr(string, '"'))
	{
		*jp='"';
		jp++; ip = string;

		/* Excel doesn't like strings begining with - */
		if (*string == '-' && *(string + 1) != (char)NULL)
			*(jp++) = '\\';

		while (*ip != (char)NULL)
		{
			if (*ip == '"')
				*jp++ = '"';
			*jp++ = *ip++;
		}
		*jp++ = '"';
		*jp = (char)NULL;
	}
	else
	{
		/* Excel doesn't like strings begining with - */
		if (*string == '-' && *(string + 1) != (char)NULL)
			strcpy(hold, "\\");
		strcat(hold, string);
	}

	if ((mode & WEB) == WEB)
	{
		strcpy(work, hold);
		jp = hold;

		if (flag == 'x')
			strcpy(hold, "<em>"), jp += 4;

		for(ip = work; *ip != (char)NULL; ip++)
		{
			if (*ip =='&')
			{
				strcpy(jp, "&amp;");
				jp += 5;
			}
			else if (*ip == '<')
			{
				strcpy(jp, "&lt;");
				jp += 4;
			}
			else if (*ip == '>')
			{
				strcpy(jp, "&gt;");
				jp += 4;
			}
			else
				*jp++ = *ip;
		}
		*jp = (char)NULL;

		if (flag == 'x')
			strcat(hold, "</em>"), jp += 4;
	}

	return hold;
}

/* this routine uses a common work area so must not be
	called twice in the same expression */
char * squash(char letter, char pad)
{	/* removed all but the first letter, if it is a null field, return pad */
	extern char hold2[];
	hold2[1] = (char)NULL;
	if (letter != '-' && letter != ' ')
		hold2[0] = letter;
	else
		hold2[0] = pad;
	return hold2;
}

/* 4.6.10 this routine uses a common work area so must not be
	called twice in the same expression */
char * sprep(char * string, char letter)
{	/* prepare numeric + alpha suffix for printing */
	char hold3[10];
	strcpy(hold3, string);
	strcat(hold3, squash(letter, (char)NULL));
	return prepare(hold3, 0);
}

int cfind(char * * ccode, char * code)
{
	int i;
	for (i = 0; ccode[i] != (char *)NULL && strcmp(ccode[i], code) != 0; i++)
		;
	if (ccode[i] == (char *)NULL)
		return -1;
	return i;
}

/* return code is passed back as return code of main */
/* 4.7.0 change reg field to mean book number */
int webhead(FILE * out, char * input, char * proref, char * book, char * parish)
{
	extern FILE * tmp;
	extern char ppath[];
	char temp[128];
	char c;

	/* 4.5.4 locate it in the binary directory */
	strcpy(temp, makepath(ppath, "template", "htm"));

	if ((tmp = fopen(temp, "r")) == (FILE *)NULL)
	{
		fprintf(err, "Fatal: Template file (%s) failed to open\n", temp);
		fprintf(stderr, "Fatal: Template file (%s) failed to open\n", temp);
		perror("Fatal");
		return 12;
	}

	while (TRUE)
	{
		c = getc(tmp);
		if (feof(tmp))
		{
			fprintf(err, "Fatal: Template file short, it ends with a lone %%\n");
			fprintf(stderr, "Fatal: Template file short, it ends with a lone %%\n");
			fclose(tmp);
			return 12;
			
		}
		if (c == '%')
		{
			c = getc(tmp);
			if (c == 'd')
				return 0;
			/* 4.7.0 change reg field to mean book number */
			escape(c, out, input, proref, book, parish);
		}
		else
			putc(c, out);
	}
}

/* 4.7.0 change reg field to mean book number */
void escape(char c, FILE * out, char * input, char * proref, char * book, char * parish)
{
	extern int year, con, ipart;
	extern char date[];

	switch (c)
	{
		case 'p':
			fputs(proref, out);
			/* 4.6.2 */
			if (ipart != 0)
				fprintf(out, "(%s)", parish);
			break;
		case 'f':
			fputs(input, out);
			break;
		case 'y':
			fprintf(out, "%d", year);
			break;
		case 'v':
			fputs(VERSION, out);
			putc(' ', out);
			fputs(date, out);
			break;
		case 'r':
			if (con)
				/* 4.7.0 change reg field to mean book number - this code is now obsolete */
				fprintf(out, " <small>COCP code %s%s</small>", proref + 6, book);
			break;
		default:
			putc(c, out);
	}

	return;
}

/* 4.7.0 change reg field to mean book number */
void webfoot(FILE * out, char * input, char * proref, char * book, char * parish)
{
	char c;

	while (TRUE)
	{
		c = getc(tmp);
		if (feof(tmp))
		{
			fclose(tmp);
			return;
			
		}
		if (c == '%')
		{
			c = getc(tmp);
			/* 4.7.0 change reg field to mean book number */
			escape(c, out, input, proref, book, parish);
		}
		else
			putc(c, out);
	}
}

void splitpath(const char * string, char * path, char * file, char * ext)
{
	char * cp;

	strcpy(path, string);
	*file = (char)NULL;
	*ext = (char)NULL;

#ifdef UNIX
	if ((cp = strrchr(path, '/')) != (char *)NULL)
#else
	if ((cp = strrchr(path, '\\')) != (char *)NULL)
#endif
	{
		*cp = (char)NULL;
		strcpy(file, cp + 1);
	}
	else
	{
		strcpy(file, path);
		*path = (char)NULL;
	}

	if ((cp = strchr(file, '.')) != (char *)NULL)
	{
		*cp = (char)NULL;
		strcpy(ext, cp + 1);
	}

	return;
}

char * makepath(const char * path, const char * file, const char * ext)
{
	extern char hold[];

	*hold = (char)NULL;
	if (*path != (char)NULL)
	{
		strcpy(hold, path);
#ifdef UNIX
		strcat(hold, "/");
#else
		strcat(hold, "\\");
#endif
	}
	strcat(hold, file);
	strcat(hold, ".");
	strcat(hold, ext);

	return hold;
}

int vote(unsigned char * buf)
{
	int i, n = 0;
	unsigned char c;

	for (i = 0; i < MAXVOTE; i++)
	{
		c = buf[vmasks[i].p];

		switch (vmasks[i].t)
		{
			case 'n':
				if (isdigit(c))
					n++;
				break;
			case 'x':
				if (c == '-' || c == 'X' || c == 'x')
					n++;
				break;
			case 'A':
				/* include 8 bit accented characters here */
				if (isupper8(c) || c == '-')
					n++;
				break;
		}
	}

	return n;
}

int maxvote(unsigned char * buf, int rlen)
{
	int i, v, im = 0, vm = 0;

	for (i = 0; i < rlen; i++)
	{
		v = vote(buf + i);
		/* bias the vote in favour of matches close to the start.
			This stops it finding the next record as perfect if the current one
			is only a little wrong. */
		if (i > 200)
	--v;
		if (v > vm)
		{
			vm = v;
			im = i;
			if (vm == MAXVOTE)
				break;
		}
	}

	return im;
}

/* test to see if a value is numeric */
int numeric(char * string)
{
	char * p = string;
	while (*p != (char)NULL)
/* v4.7.3	it must be a bug in my library but the 1/2 (189) and
		3/4 (190) but not 1/4 (188) symbols are treated as
		digits!
		if (! isdigit(*p++)) */
		if (*p++ < '0' || *p > '9')
			return FALSE;
	return TRUE;
}

/* capitalise a string (7 bit as it is not used for data) */
char * capitals(char * string)
{
	int i = 0;

	while (string[i] != (char)NULL)
		string[i++] = toupper(string[i]);

	return string;
}

/* return code is passed back as return code of main */
int decode_census(char * fname, char * webname, char * proref)
{
	extern int year, sct, con, ipart;
	extern FILE * err;
	/* 4.5.3 extend all path and file names to 128 (XP likes long paths) */
	char Fname[128], class[15], piece[10], padpiece[10], part[3], work[20];
	int i;

	con = sct = FALSE;

	/* take a copy in capitals to make it easier */
	strcpy(Fname, fname);
	(void)capitals(Fname);
	strcpy(part, "00");

	/* This could be more compact, but is like this for clarity */
	if (strncmp(Fname, "HO1", 3) == 0) /* Cornwall & Original 1841 */
	{
		year = 1841;
		strcpy(fname, "ho1");
		/* this is a fiddle to allow CON to use the new part numbering scheme
			and also caters for the old HO107 scheme */
		if (Fname[4] > '6')
		{
			con = TRUE;
			/* The problem is that we had already divided the pieces into more
			than 9 parts so we use the next digit as well. The sequence
			is 17 ... 97, 18 ... 98, 19 ... 99  (08 and 09 are not used )
			so this allows for 27 parts. It works because the piece numbers
			are only three digits. 07 caters for the old HO107 pattern */
			strncpy(part, Fname + 3, 2);
			ipart = (part[1] - '7') * 10 + part[0] - '0';
			if (ipart == 10 || ipart == 20)
			{
				fprintf(err,
					"Fatal: Part number in file name (%s) is not valid\n",
					part);
				fprintf(stderr,
					"Fatal: Part number in file name is not valid\n");
				return 12;
			}
			/* fill in the gaps not used */
			if (ipart > 20) --ipart;
			if (ipart > 10) --ipart;
			strcpy(piece, Fname + 5);
			*webname = (char)NULL;

			/* 4.6.0 revert original HO107 to be not CON */
			if (ipart == 0)
			{
				con = FALSE;
				strcpy(fname, "ho107");
				strcpy(part, "00");
				strcpy(webname, "40");
			}
		}
		else /* 4.6.0 the next allows for HO10nnnn which was allowed for a
			short while by accident - I don't think it was ever used */
		{	strcat(fname, "0");
			strcpy(piece, Fname + 4);
			strcpy(webname, "4");
		}
		strcpy(class, "HO107");
	}
	/* v 4.6.0 allow proper ENG 1841 with 26 sub parts */
	else if (strncmp(Fname, "HO4", 3) == 0) /* England/Wales 1841 */
	{
		year = 1841;
		strcpy(fname, "ho41");
		/* Allow 26 sub parts */
		if (Fname[3] >= 'A' && Fname[3] <= 'Z')
		{
			sprintf(part, "%0.2d", Fname[3] - 'A' + 1);
			fname[3] = 'a' + Fname[3] - 'A';
		}
		strcpy(piece, Fname + 4);
		strcpy(class, "HO107");
		strcpy(webname, "4");
	}
	else if (strncmp(Fname, "HS4", 3) == 0) /* Scotland 1841 */
	{
		year = 1841;
		sct = TRUE;
		strcpy(fname, "hs4");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1841"); /* bogus */
		strcpy(webname, "S4");
	}
	else if (strncmp(Fname, "HO5", 3) == 0) /* England/Wales 1851 */
	{
		year = 1851;
		strcpy(fname, "ho51");
		/* Allow 26 sub parts */
		if (Fname[3] >= 'A' && Fname[3] <= 'Z')
		{
			sprintf(part, "%0.2d", Fname[3] - 'A' + 1);
			fname[3] = 'a' + Fname[3] - 'A';
		}
		strcpy(piece, Fname + 4);
		strcpy(class, "HO107");
		strcpy(webname, "5");
	}
	else if (strncmp(Fname, "HS5", 3) == 0) /* Scotland 1851 */
	{
		year = 1851;
		sct = TRUE;
		strcpy(fname, "hs5");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1851"); /* bogus */
		strcpy(webname, "S5");
	}
	else if (strncmp(Fname, "RG09", 4) == 0) /* England/Wales 1861 */
	{
		year = 1861;
		strcpy(fname, "rg09");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG9");
		strcpy(webname, "6");
	}
	else if (strncmp(Fname, "RG9", 3) == 0) /* just in case */
	{
		year = 1861;
		strcpy(fname, "rg09");
		strcpy(piece, Fname + 3);
		strcpy(class, "RG9");
		strcpy(webname, "6");
	}
	else if (strncmp(Fname, "RS6", 3) == 0) /* Scotland 1861 */
	{
		year = 1861;
		sct = TRUE;
		strcpy(fname, "rs6");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1861"); /* bogus */
		strcpy(webname, "S6");
	}
	else if (strncmp(Fname, "RG10", 4) == 0) /* England/Wales 1871 */
	{
		year = 1871;
		strcpy(fname, "rg10");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG10");
		strcpy(webname, "7");
	}
	else if (strncmp(Fname, "RS7", 3) == 0) /* Scotland 1871 */
	{
		year = 1871;
		sct = TRUE;
		strcpy(fname, "rs7");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1871"); /* bogus */
		strcpy(webname, "S7");
	}
	else if (strncmp(Fname, "RG11", 4) == 0) /* England/Wales 1881 */
	{
		year = 1881;
		strcpy(fname, "rg11");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG11");
		strcpy(webname, "8");
	}
	else if (strncmp(Fname, "RS8", 3) == 0) /* Scotland 1881 */
	{
		year = 1881;
		sct = TRUE;
		strcpy(fname, "rs8");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1881"); /* bogus */
		strcpy(webname, "S8");
	}
	else if (strncmp(Fname, "RG12", 4) == 0) /* England/Wales 1891 */
	{
		year = 1891;
		strcpy(fname, "rg12");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG12");
		*webname = (char)NULL;
	}
	else if (strncmp(Fname, "RS9", 3) == 0) /* Scotland 1891 */
	{
		year = 1891;
		sct = TRUE;
		strcpy(fname, "rs9");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1891"); /* bogus */
		strcpy(webname, "S9");
	}
	else if (strncmp(Fname, "RG13", 4) == 0) /* England/Wales 1901 */
	{
		year = 1901;
		strcpy(fname, "rg13");
		strcpy(piece, Fname + 4);
		strcpy(class, "RG13");
		strcpy(webname, "0");
	}
	else if (strncmp(Fname, "RS0", 3) == 0) /* Scotland 1901 */
	{
		year = 1901;
		sct = TRUE;
		strcpy(fname, "rs0");
		strncpy(part, Fname + 3, 2);
		strcpy(piece, Fname + 5);
		strcpy(class, "SCT 1901"); /* bogus */
		strcpy(webname, "S0");
	}
	else
	{
		fprintf(err, "Fatal: Census year cannot be determined from file name (%s)\n", fname);
		fprintf(stderr, "Fatal: Census year cannot be determined from file name\n");
		return 12;
	}

	/* check that part number is numeric */
	if (! numeric(part))
	{
		fprintf(err, "Fatal: Part number in file name (%s) is not numeric\n",
			part);
		fprintf(stderr, "Fatal: Part number in file name is not numeric\n");
		return 12;
	}

	/* this is already done for CON 1841 which uses a different algorithm */
	if (! con)
		ipart = atoi(part);

	if (sct)
	{
		if (ipart < 10)
		{
			fprintf(err, "Fatal: Part number in file name (%s) is not valid\n",
				part);
			fprintf(stderr, "Fatal: Part number in file name is not valid\n");
			return 12;
		}
		ipart -= 10;
	}

	/* check that piece number is numeric */
	if (! numeric(piece))
	{
		fprintf(err, "Fatal: Piece number in file name (%s) is not numeric\n",
			piece);
		fprintf(stderr, "Fatal: Piece number in file name is not numeric\n");
		return 12;
	}

	/* 4.6.0 this section all juggled around a bit */

	/* put the part number into the file name */
	if (con || sct)
		strcat(fname, part);

	/* tidy up piece number */
	i = atoi(piece);
	sprintf(piece, "%d", i);

	if (strlen(fname) == 5)
		sprintf(padpiece, "%0.3d", i);
	else
		sprintf(padpiece, "%0.4d", i);

	/* finish off file names */
	strcat(fname, padpiece);

	/* create PRO reference number */
	strcpy(proref, class);
	strcat(proref, "/");
	strcat(proref, piece);

	/* create webname (HTML file name) */
	strcat(webname, padpiece);
	if (!con && ipart != 0)
	{
		strcat(webname, "a");
		webname[5] = 'a' - 1 + ipart;
	}

	/* CON 1841 webname still needs parish number which will be added later */	

	return 0;
}

char decode_emp(char * occ)
{
	char emp;
	char * cp;
	extern char hold[];

	strcpy(hold, occ);
	(void)capitals(hold);

	/* these codes always were a daft idea, ambiguous, hard to type, took up too much
		space. A separate column would have been much better */
	/* We don't check for the trailing bracket */
	/* The code must be the last thing on the line (because anything later is lost) */
	if ((cp = strstr(hold, "(EM'EE")) != (char *)NULL)
		emp='E';
	else if ((cp = strstr(hold, "(NOTEM")) != (char *)NULL)
		emp='N';
	else if ((cp = strstr(hold, "(EM'ER")) != (char *)NULL)
		emp='R';
	/* the rest are malformed cases */
	else if ((cp = strstr(hold, "(EMP'EE")) != (char *)NULL)
		emp='E';
	else if ((cp = strstr(hold, "(EMEE")) != (char *)NULL)
		emp='E';
	/* (EM'E and (EMP'E are not checked because they are ambiguous */
	else if ((cp = strstr(hold, "(NOT E")) != (char *)NULL)
		emp='N';
	else if ((cp = strstr(hold, "(NOT'E")) != (char *)NULL)
		emp='N';
	/* (NOTE could mean Note: so we don't check for it */
	else if ((cp = strstr(hold, "(EMP'ER")) != (char *)NULL)
		emp='R';
	else if ((cp = strstr(hold, "(EMER")) != (char *)NULL)
		emp='R';
	else if ((cp = strstr(hold, "(EM'R")) != (char *)NULL)
		emp='R';
	/* Allow for a leading space */
	else if ((cp = strstr(hold, "( EM'EE")) != (char *)NULL)
		emp='E';
	else if ((cp = strstr(hold, "( EMP'EE")) != (char *)NULL)
		emp='E';
	else if ((cp = strstr(hold, "( EMEE")) != (char *)NULL)
		emp='E';
	else if ((cp = strstr(hold, "( NOTEM")) != (char *)NULL)
		emp='N';
	else if ((cp = strstr(hold, "( NOT E")) != (char *)NULL)
		emp='N';
	else if ((cp = strstr(hold, "( NOT'E")) != (char *)NULL)
		emp='N';
	else if ((cp = strstr(hold, "( EM'ER")) != (char *)NULL)
		emp='R';
	else if ((cp = strstr(hold, "( EMP'ER")) != (char *)NULL)
		emp='R';
	else if ((cp = strstr(hold, "( EMER")) != (char *)NULL)
		emp='R';
	else if ((cp = strstr(hold, "( EM'R")) != (char *)NULL)
		emp='R';
	/* I am told that there are cases with {} and [] but I don't check for those yet.
		It could be done by conversion in capitals() */
	else
		emp = (char)NULL;

	if (emp != (char)NULL)
	{
		cp = occ + (cp - hold);
		*cp =(char)NULL;

		/* check for the last character now being a space */
		if (cp != occ && *(cp - 1) == ' ')
			*(--cp) = (char)NULL;

		/* check for the last character now being a comma */
		if (cp != occ && *(cp - 1) == ',')
			*(--cp) = (char)NULL;
	}

	return emp;
}

/* 8 bit version of isupper to allow for accented characters ISO8859-1 */
int isupper8(unsigned char c)
{
	/*         A  to       Z           Agrave   to  THORN   not multiply */
	if ((c >= 'A' && c <= 'Z') || (c >= 0xC0 && c <= 0xDE && c != 0xD7))
		return TRUE;
	return FALSE;
}

/* 8 bit version of tolower to allow for accented characters ISO8859-1 */
unsigned char tolower8(unsigned char c)
{
	/*         A  to       Z           Agrave   to  THORN   not multiply */
	if ((c >= 'A' && c <= 'Z') || (c >= 0xC0 && c <= 0xDE && c != 0xD7))
		return c + 0x20;
	return c;
}
