/* fixparms */
#define VERSION "1.3.1"

/* Uncomment the next line to compile on a unix system */
/* #define UNIX */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

void usage(void)
{
	fprintf(stderr,"Usage (v%s): fixparms [-a] [-h|-?] <filename.DAT>\n", VERSION);
	fprintf(stderr,"\ta=ALL, h|?=HELP\n");
}

/* Copyright (c) 2002 Rick Parsons.

	This software is published Open Source Freeware. You are free to use the
	programs either in full or part without charge. I would ask however that
	this notice is included with any extract or onward distribution.

	No charge may be made for the use of this code, however there is no
	objection to including modest extracts (with acknowledgement) in otherwise
	commercial software.

	No warranty is given or implied by the use of the software which you do at
	your own risk. No compensation can be considered regarding damage to data,
	computers or any thing else arising as a result of using these programs.
*/

#define TRUE 1
#define FALSE 0

#define h_trid 0
#define h_unk 4
#define h_name 8

#define a_piece 0
#define a_null 4
#define a_reg 8
#define a_lds 28
#define a_null2 36
#define a_house 38
#define a_rec 44
#define a_null3 50
#define a_a 63

#define b_piece 0
#define b_null 4
#define b_parish 8
#define b_null2 28
#define b_spc 33
#define b_star 34
#define b_trid 35
#define b_date 39
#define b_chid 48
#define b_chdate 52
#define b_null3 61
#define b_b 63

int options(int, char **);
void strip(char *);
char * pad(char *, int);
void splitpath(const char *, char *, char *, char *);
char * makepath(const char *, const char *, const char *);
char * capitals(char *);

static int all;
static char hold[100];

int _main(int argc, char *argv[])
{
	int i;
	char buf[65];
	char file[64];
	char fname[15];
	char path[64];
	char ext[10];
	char output[64];
	char fpiece[10], fnum[60]; /* room for a bit of error */
	FILE * f;
	FILE * out;
	time_t t;
	struct tm * tp;

	int hn, rn, n, update;
		char trid[5];
		char trid_p[5];
		char unk[5];
		char name[57];
		char piece[5];
		char reg[21];
		char lds[9];
		char house[7];
		char rec[7];
		char parish[21];
		char date[10];
		char date_p[10];

		/* preset all trailing nulls */

		trid[4] = (char) NULL;
		trid_p[4] = (char) NULL;
		unk[4] = (char) NULL;
		name[56] = (char) NULL;
		piece[4] = (char) NULL;
		reg[20] = (char) NULL;
		lds[8] = (char) NULL;
		house[6] = (char) NULL;
		rec[6] = (char) NULL;
		parish[20] = (char) NULL;
		date_p[9] = (char) NULL;

	if (argc == 1)
	{
		/* If no arguments we call the usage routine and exit */
		usage();
		return 1;
	}

	/* handle the program options */
	i = options(argc,argv);

	/* If there is still no file name then what do we do? */
	if (i == 0)
		return 2;

		strcpy(file, argv[i]);
		if ((f = fopen(file, "rb+")) == (FILE *)NULL)
		{
			fprintf(stderr, "Fatal: Data file %s failed to open\n", file);
			perror("Fatal");
			return 12;
		}
		setvbuf(f, (char *)NULL, _IONBF, 0); /* no buffering */

		/* get today's date */
		t = time((time_t *)NULL);
		tp = localtime(&t);
		strftime(date, 10, "%d%b%Y", tp);

		i = fread(buf, sizeof(char), 64, f);
		if (feof(f) || i != 64)
		{
			fprintf(stderr, "Fatal: No header record\n");
			fclose(f);
			return 16;
		}

		/* now decode record */
		strncpy(trid, buf + h_trid, 4); /* strip(trid); */
		strncpy(unk, buf + h_unk, 4); /* strip(unk); */
		strncpy(name, buf + h_name, 56); strip(name);

	if (all)
	{
		splitpath(file, path, fname, ext);
		strcpy(output, makepath(path, fname, "lst"));
		if ((out = fopen(output, "w")) == (FILE *)NULL)
		{
			fprintf(stderr, "Fatal: Output file %s failed to open\n", output);
			perror("Fatal");
			return 12;
		}

		fprintf(out, "Listing for fixparms v%s (Rick Parsons) listing %s\n"
			"[When reporting program errors, be sure to send this file and the original data file]\n\n",
			VERSION, file);

		/* header */
		fprintf(out, "Transcriber id : %s, Name : %s\n", trid, name);
		n = 1;

		/* read in the first piece record */
		i = fread(buf, sizeof(char), 64, f);

		while (TRUE)
		{
			if (feof(f))
				break;

			n++;
			if (i != 64)
			{
				fprintf(out, "Error: record %d short - damaged file\n", n);
				break;
			}

			/* piece record */
			strncpy(piece, buf + a_piece, 4); /* strip(piece); */
			strncpy(reg, buf + a_reg, 20); strip(reg);
			strncpy(lds, buf + a_lds, 8); strip(lds);
			strncpy(house, buf + a_house, 6); /* strip(house); */
			hn = atoi(house);
			strncpy(rec, buf + a_rec, 6); /* strip(rec); */
			rn = atoi(rec);

			fprintf(out, "\nPiece %s\n", piece);
			fprintf(out, "Registration District: %s, LDS Film: %s\n", reg, lds);
			fprintf(out, "Current household count: %d, Record count: %d\n\n", hn, rn);

			strcpy(fpiece, piece);

			while (TRUE)
			{
				i = fread(buf, sizeof(char), 64, f);

				if (feof(f))
					break;

				n++;
				if (i != 64)
				{
					fprintf(out, "Error: record %d short - damaged file\n", n);
					break;
				}

				/* decode piece number */
				strncpy(piece, buf + b_piece, 4); /* strip(piece); */
				if (strcmp(fpiece, piece))
					break;
					/* loop around (we have already read in the first record of the
               	next piece */

				/* now decode rest of record */
				strncpy(parish, buf + b_parish, 20); strip(parish);

				fprintf(out, "Civil parish: %s, ", parish);

				if (buf[b_star] == '*' || buf[b_star] == 'C')
				{
					strncpy(trid_p, buf + b_trid, 4); /* strip(trid_p); */
					strncpy(date_p, buf + b_date, 9); /* strip(date_p); */
					fprintf(out, "Updated by: %s On %s\n", trid_p, date_p);
				}
				else
					fprintf(out, "Not updated\n");
			}
		}

		fprintf(stderr, "\nOutput file is %s\n", output);

		fclose(out);
		fclose(f);

		return 0;
	}

		fprintf(stderr, "Transcriber id : %s, Name : %s\n", trid, name);

		update = FALSE;
		fprintf(stderr, "\n\nEnter New Transcriber id (RETURN to leave unchanged): ");
		gets(fnum);
		if (strlen(fnum) > 0)
			strcpy(trid, capitals(fnum)), update = TRUE;
		fprintf(stderr, "\n\nEnter New Name (RETURN to leave unchanged): ");
		gets(fnum);
		if (strlen(fnum) > 0)
			strcpy(name, fnum), update = TRUE;
		if (update)
		{
			/* seek to start of the file */
			fseek(f, 0, SEEK_SET);
			memset(buf, ' ', 64);
			strncpy(buf + h_trid, trid, strlen(trid));
			strncpy(buf + h_name, name, strlen(name));
			buf[64] = (char)NULL;
			fprintf(f, "%s", buf);
		}

	while (TRUE) /* loop around asking for piece numbers */
	{
		fprintf(stderr, "\nEnter Piece Number (4 digits, RETURN to end): ");
		gets(fpiece);
		if (strlen(fpiece) == 0)
			break;
		fseek(f, 64, SEEK_SET);
		n = 0;

	do
	{
		i = fread(buf, sizeof(char), 64, f);
		if (feof(f))
			break;
		if (i != 64)
		{
			fprintf(stderr, "Error: Short record - damaged file\n");
			fclose(f);
			return 20;
		}
		n++;

		/* decode piece number */
		strncpy(piece, buf + a_piece, 4); /* strip(piece); */
	} while (strcmp(fpiece, piece));

	if (feof(f))
	{
		fprintf(stderr, "Error: Piece %s not found\n", fpiece);
		continue;
	}

		/* now decode rest of record */
		strncpy(reg, buf + a_reg, 20); strip(reg);
		strncpy(lds, buf + a_lds, 8); strip(lds);
		strncpy(house, buf + a_house, 6); /* strip(house); */
		hn = atoi(house);
		strncpy(rec, buf + a_rec, 6); /* strip(rec); */
		rn = atoi(rec);

		fprintf(stderr, "Registration District: %s, LDS Film: %s\n", reg, lds);
		fprintf(stderr, "Current household count: %d, Record count: %d", hn, rn);

		update = FALSE;
		fprintf(stderr, "\n\nEnter New household count (RETURN to leave unchanged): ");
		gets(fnum);
		if (strlen(fnum) > 0)
			hn = atoi(fnum), update=TRUE;
		fprintf(stderr, "Enter New record count (RETURN to leave unchanged): ");
		gets(fnum);
		if (strlen(fnum) > 0)
			rn = atoi(fnum), update = TRUE;
		if (update)
		{
			/* seek to exactly the right place in the file */
			fseek(f, n * 64 + a_house, SEEK_SET);
			fprintf(f, "%0.6d%0.6d", hn, rn);
		}

		/* now have a look at the parish records */
		fseek(f, (n + 1) * 64, SEEK_SET);
		while (TRUE)
		{
			i = fread(buf, sizeof(char), 64, f);
			if (feof(f))
				break;
			if (i != 64)
			{
				fprintf(stderr, "Error: Short record - damaged file\n");
				fclose(f);
				return 20;
			}
			n++;

			/* decode piece number */
			strncpy(piece, buf + b_piece, 4); /* strip(piece); */
			if (strcmp(fpiece, piece))
				break;
			/* now decode rest of record */
			strncpy(parish, buf + b_parish, 20); strip(parish);

			fprintf(stderr, "Civil parish: %s, ", parish);

			if (buf[b_star] == '*' || buf[b_star] == 'C')
			{
				strncpy(trid_p, buf + b_trid, 4); /* strip(trid_p); */
				strncpy(date_p, buf + b_date, 9); /* strip(date_p); */
				fprintf(stderr, "Updated by: %s On %s\n", trid_p, date_p);

				fprintf(stderr, "\nEnter Y to clear, RETURN to leave unchanged): ");
				gets(fnum);
				if (strlen(fnum) == 1 && (fnum[0] == 'Y' || fnum[0] == 'y'))
				{
					/* seek to exactly the right place in the file */
					fseek(f, n * 64 + b_star, SEEK_SET);
					for (i = 0; i < (b_null3 - b_star); i++)
						fputc((char)NULL, f);
					fseek(f, (n + 1) * 64, SEEK_SET);
				}
			}
			else
			{
				fprintf(stderr, "Not updated\n");

				fprintf(stderr, "\nEnter Y to set, RETURN to leave unchanged): ");
				gets(fnum);
				if (strlen(fnum) == 1 && (fnum[0] == 'Y' || fnum[0] == 'y'))
				{
					/* seek to exactly the right place in the file */
					fseek(f, n * 64 + b_star, SEEK_SET);
					fputc('*', f);
					fputs(trid, f);
					fputs(date, f);
					fseek(f, (n + 1) * 64, SEEK_SET);
				}
			}
		}

	} /* loop around to next piece number */

	fclose(f);

	fprintf(stderr, "\nFixparms finished\n");

	return 0;
}

void strip(char * string)
{
	int l, i;
	l = strlen(string);
	for(i = --l; i >= 0; --i)
		if (string[i] != ' ')
			break;
	string[i + 1] = (char) NULL;
	return;
}

char * pad(char * string, int len)
{
	char * p;
	int i, l;
	l = strlen(string);
	p = string + l;
	for (i = len - l; i > 0; --i)
		*(p++) = ' ';
	*p = (char)NULL;

	return string;
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash (unix) or a slash (DOS) */
int options(int argc,char *argv[])
{
	int i,firstnonoption=0;
	extern int mode, rlen, verbose;

	all = FALSE;

	for (i=1; i< argc;i++)
	{
		if (
#ifndef UNIX
			argv[i][0] == '/' ||
#endif
			argv[i][0] == '-')
		{
			switch (argv[i][1])
			{
				case 'a':
				case 'A':
					all = TRUE;
					break;
				case 'h':
				case 'H':
				case '?':
					usage();
					return 0;
				case 'i':
				case 'I':
					all = FALSE;
					break;
				default:
					fprintf(stderr,"Unknown option %s\n",argv[i]);
					break;
			}
		}
		else
		{
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

void splitpath(const char * string, char * path, char * file, char * ext)
{
	char * cp;

	strcpy(path, string);
	*file = (char)NULL;
	*ext = (char)NULL;

#ifdef UNIX
	if ((cp = strrchr(path, '/')) != (char *)NULL)
#else
	if ((cp = strrchr(path, '\\')) != (char *)NULL)
#endif
	{
		*cp = (char)NULL;
		strcpy(file, cp + 1);
	}
	else
	{
		strcpy(file, path);
		*path = (char)NULL;
	}

	if ((cp = strchr(file, '.')) != (char *)NULL)
	{
		*cp = (char)NULL;
		strcpy(ext, cp + 1);
	}

	return;
}

char * makepath(const char * path, const char * file, const char * ext)
{
	extern char hold[];

	*hold = (char)NULL;
	if (*path != (char)NULL)
	{
		strcpy(hold, path);
#ifdef UNIX
		strcat(hold, "/");
#else
		strcat(hold, "\\");
#endif
	}
	strcat(hold, file);
	strcat(hold, ".");
	strcat(hold, ext);

	return hold;
}

/* capitalise a string */
char * capitals(char * string)
{
	int i = 0;

	while (string[i] != (char)NULL)
		string[i++] = toupper(string[i]);

	return string;
}

